from pathlib import Path

# we should make this a proper package that loads examples
# similar how they handle it in xarray:
# https://github.com/pydata/xarray/blob/main/xarray/tutorial.py
# Currently there are simple paths to the examples which are then easily loaded in the notebooks

power_grid_model = Path(__file__).parent / "power-grid-example.xlsx"
simple_example = Path(__file__).parent / "sheet2cdf-transformation-rules.xlsx"
source_to_solution_mapping = Path(__file__).parent / "source-to-solution-mapping-rules.xlsx"



import logging
from pathlib import Path
from fastapi import APIRouter

from cognite.neat.rules.parser import parse_rules_from_excel_file
from cognite.neat.workflows.utils import get_file_hash

from cognite.neat.app.api.configuration import neat_app


router = APIRouter()


@router.get("/api/rules")
def get_rules(
    sheetname: str = "Properties",
    url: str | None = None,
    source_type: str | None = None,
    orient: str = "columns",
    workflow_name: str = "default",
    file_name: str | None = None,
    version: str | None = None,
):
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    if not file_name:
        version = workflow.get_config_item("rules.version").value
        file_name = workflow.get_config_item("rules.file").value
    path = Path(neat_app.config.rules_store_path, file_name)
    src = "local"
    if url:
        path = url

    if path.exists() and not version:
        logging.info(f"Loading rules from {path}")
    elif path.exists() and version:
        hash = get_file_hash(path)
        if hash != version:
            neat_app.cdf_store.load_rules_file_from_cdf(file_name, version)
            src = "cdf"
    else:
        neat_app.cdf_store.load_rules_file_from_cdf(file_name, version)
        src = "cdf"

    error_text = ""
    properties = []
    classes = []
    try:
        rules = parse_rules_from_excel_file(path)
        properties = [
            {
                "class": value.class_id,
                "property": value.property_id,
                "property_description": value.description,
                "property_type": value.expected_value_type,
                "cdf_resource_type": value.cdf_resource_type,
                "cdf_metadata_type": value.resource_type_property,
                "rule_type": value.rule_type,
                "rule": value.rule,
            }
            for value in rules.properties.values()
        ]

        classes = [
            {
                "class": value.class_id,
                "class_description": value.description,
                "cdf_resource_type": value.cdf_resource_type,
                "cdf_parent_resource": value.parent_asset,
            }
            for value in rules.classes.values()
        ]
    except Exception as e:
        error_text = str(e)

    return {
        "properties": properties,
        "classes": classes,
        "file_name": path.name,
        "hash": get_file_hash(path),
        "error_text": error_text,
        "src": src,
    }



"""This module contains the definition of validation errors and warnings raised when parsing transformation rules.
CODES:
- 0 - 99: reserved for errors and warnings related to general errors/warnings
- 100 - 199: reserved for Metadata sheet
- 200 - 299: reserved for Classes sheet
- 300 - 399: reserved for Properties sheet
- 400 - 499: reserved for Prefixes sheet
- 500 - 599: reserved for Instances sheet
- 600 - 699: reserved for Transformation Rules, usually checking inter-sheet dependencies
"""

#########################################
# Metadata sheet Error Codes 100 - 199: #

from typing import Any
from warnings import WarningMessage

from pydantic_core import ErrorDetails, PydanticCustomError
from rdflib import URIRef


def wrangle_warnings(list_of_warnings: list[WarningMessage]) -> list[dict]:
    warning_list: list[dict] = []
    for warning in list_of_warnings:
        if issubclass(warning.message.__class__, NeatWarning):
            warning_list.append(_neat_warning_to_dict(warning))
        elif issubclass(warning.message.__class__, Warning):
            warning_list.append(_python_warning_to_dict(warning))
    return warning_list


def _neat_warning_to_dict(warning: WarningMessage) -> dict:
    return {
        "type": warning.category.type_,
        "loc": (),
        "msg": str(warning.message),
        "input": None,
        "ctx": dict(
            type_=warning.category.type_,
            code=warning.category.code,
            description=warning.category.description,
            example=warning.category.example,
            fix=warning.category.fix,
        ),
    }


def _python_warning_to_dict(warning: WarningMessage) -> dict:
    return {
        "type": warning.category,
        "loc": (),
        "msg": str(warning.message),
        "input": None,
        "ctx": dict(
            type_=warning.category,
            code=None,
            description=None,
            example=None,
            fix=None,
        ),
    }


class NeatError(Exception):
    type_: str
    code: int
    description: str
    example: str
    fix: str
    message: str

    def to_pydantic_custom_error(self):
        return PydanticCustomError(
            self.type_,
            self.message,
            dict(type_=self.type_, code=self.code, description=self.description, example=self.example, fix=self.fix),
        )

    def to_error_dict(self) -> ErrorDetails:
        return {
            "type": self.type_,
            "loc": (),
            "msg": self.message,
            "input": None,
            "ctx": dict(
                type_=self.type_,
                code=self.code,
                description=self.description,
                example=self.example,
                fix=self.fix,
            ),
        }


class NeatWarning(UserWarning):
    type_: str
    code: int
    description: str
    example: str
    fix: str
    message: str


class Error0(NeatError):
    type_: str = "ExcelFileMissingMandatorySheets"
    code: int = 0
    description: str = "Given Excel file is missing mandatory sheets"
    example: str = ""
    fix: str = ""

    def __init__(self, missing_sheets: set[str], verbose=False):
        self.missing_fields = missing_sheets

        self.message = (
            "Given Excel file is not compliant Transformation Rules file."
            f" It is missing mandatory sheets: {', '.join(missing_sheets)}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error1(NeatError):
    type_: str = "NotValidRDFPath"
    code: int = 1
    description: str = "Provided rdf path is not valid, i.e. it cannot be converted to SPARQL query"
    example: str = ""
    fix: str = "Get familiar with RDF paths and check if provided path is valid"

    def __init__(self, rdf_path, verbose=False):
        self.rdf_path = rdf_path

        self.message = f"{self.rdf_path} is not a valid rdfpath!"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error2(NeatError):
    type_: str = "NotValidTableLookUp"
    code: int = 2
    description: str = "Provided table lookup is not valid, i.e. it cannot be converted to CDF lookup"
    example: str = ""
    fix: str = "Get familiar with RAW look up and RDF paths and check if provided rawlookup is valid"

    def __init__(self, table_look_up, verbose=False):
        self.table_look_up = table_look_up

        self.message = f"{self.table_look_up} is not a valid table lookup"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error3(NeatError):
    type_: str = "NotValidRAWLookUp"
    code: int = 3
    description: str = "Provided rawlookup is not valid, i.e. it cannot be converted to SPARQL query and CDF lookup"
    example: str = ""
    fix: str = "Get familiar with RAW look up and RDF paths and check if provided rawlookup is valid"

    def __init__(self, raw_look_up, verbose=False):
        self.raw_look_up = raw_look_up

        self.message = f"Invalid rawlookup expected traversal | table lookup, got {raw_look_up}"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error10(NeatError):
    type_: str = "EntitiesContainNonDMSCompliantCharacters"
    code: int = 10
    description: str = (
        "This error is raised during export of Transformation Rules to"
        " DMS schema when entities contain non DMS compliant characters."
    )
    example: str = ""
    fix: str = "Make sure to check validation report of Transformation Rules and fix DMS related warnings."

    def __init__(self, report: str = "", verbose=False):
        self.message = f"Following entities contain non DMS compliant characters: {report}"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error11(NeatError):
    type_: str = "PropertiesDefinedMultipleTimes"
    code: int = 11
    description: str = (
        "This error is raised during export of Transformation Rules to "
        "DMS schema when properties are defined multiple times for the same class."
    )
    example: str = ""
    fix: str = "Make sure to check validation report of Transformation Rules and fix DMS related warnings."

    def __init__(self, report: str = "", verbose=False):
        self.message = f"Following properties defined multiple times for the same class(es): {report}"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error20(NeatError):
    type_: str = "UnableToDownloadExcelFile"
    code: int = 11
    description: str = (
        "This error is raised during loading of byte representation of"
        " a Excel file from Github when given file cannot be downloaded."
    )
    example: str = ""
    fix: str = "Make sure you provided correct parameters to download Excel file from github repository"

    def __init__(self, filepath: str, loc: str, reason: str, verbose=False):
        self.message = f"File '{filepath}' from '{loc}' cannot be downloaded! Reason: {reason}"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error21(NeatError):
    type_: str = "NotExcelFile"
    code: int = 11
    description: str = (
        "This error is raised during loading of byte representation of a file from Github"
        " into openpyxl workbook object in case when byte representation is not Excel file."
    )
    example: str = ""
    fix: str = "Make sure you that byte representation of a file is Excel file!"

    def __init__(self, filepath: str, loc: str, verbose=False):
        self.message = f"File '{filepath}' from '{loc}' is not a valid excel file!"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error30(NeatError):
    type_: str = "PropertyDefinitionsNotForSameProperty"
    code: int = 30
    description: str = "This error is raised if property definitions are not for linked to the same property id"
    example: str = ""
    fix: str = ""

    def __init__(self, verbose=False):
        self.message = "All definitions should have the same property_id! Aborting."

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error40(NeatError):
    type_: str = "FieldValueOfUnknownType"
    code: int = 40
    description: str = (
        "This error is raised when generating in-memory pydantic model"
        " from Transformation Rules from model, when field definitions are not"
        " provided as dictionary of field names ('str') and their types ('tuple' or 'dict')."
    )
    example: str = ""
    fix: str = ""

    def __init__(self, field: str, definition: Any, verbose=False):
        self.message = (
            f"Field {field} has definition of type {type(definition)}"
            " which is not acceptable! Only definition in form of dict or tuple is acceptable!"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error41(NeatError):
    type_: str = "FieldRequiredButNotProvided"
    code: int = 41
    description: str = (
        "This error is raised when instantiating in-memory pydantic model"
        " from graph class instance which is missing required field (i.e., property)."
    )
    example: str = ""
    fix: str = "Either make field optional or add missing property to graph instance."

    def __init__(self, field: str, id_: str | URIRef, verbose=False):
        self.message = f"Field {field} is not present in graph instance {id_}!"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error51(NeatError):
    type_: str = "MetadataSheetMissingMandatoryFields"
    code: int = 51
    description: str = "Metadata sheet, which is part of Transformation Rules Excel file, is missing mandatory rows"
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose=False):
        self.missing_fields = missing_fields

        self.message = f"Metadata sheet is missing following mandatory fields: {', '.join(missing_fields)}"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error52(NeatError):
    type_: str = "ClassesSheetMissingMandatoryColumns"
    code: int = 52
    description: str = (
        "Classes sheet, which is a mandatory part of Transformation Rules Excel file, "
        "is missing mandatory columns at row 2"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose=False):
        self.missing_fields = missing_fields

        self.message = f"Classes sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 2"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error53(NeatError):
    type_: str = "PropertiesSheetMissingMandatoryColumns"
    code: int = 53
    description: str = (
        "Properties sheet, which is a mandatory part of Transformation Rules Excel file, "
        "is missing mandatory columns at row 2"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose=False):
        self.missing_fields = missing_fields

        self.message = f"Properties sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 2"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error54(NeatError):
    type_: str = "PrefixesSheetMissingMandatoryColumns"
    code: int = 54
    description: str = (
        "Prefixes sheet, which is part of Transformation Rules Excel file, is missing mandatory columns at row 1"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose=False):
        self.missing_fields = missing_fields

        self.message = f"Prefixes sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 1"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error55(NeatError):
    type_: str = "InstancesSheetMissingMandatoryColumns"
    code: int = 55
    description: str = (
        "Instances sheet, which is part of Transformation Rules Excel file, is missing mandatory columns at row 1"
    )
    example: str = ""
    fix: str = ""

    def __init__(self, missing_fields: set[str], verbose=False):
        self.missing_fields = missing_fields

        self.message = f"Instances sheet is missing following mandatory columns: {', '.join(missing_fields)} at row 1"

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


# Metadata sheet Error and Warning Codes 100 - 199:
class Error100(NeatError):
    type_: str = "PrefixRegexViolation"
    code: int = 100
    description: str = "Prefix, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If prefix is set to 'power grid', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if prefix in the 'Metadata' sheet contains any illegal characters and respects the regex expression"
    )

    def __init__(self, prefix, regex_expression, verbose=False):
        self.prefix = prefix
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid prefix '{self.prefix}' stored in 'Metadata' sheet, it must obey regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error101(NeatError):
    type_: str = "CDFSpaceRegexViolation"
    code: int = 101
    description: str = "cdfSpaceName, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If cdfSpaceName is set to 'power grid', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if cdfSpaceName in the 'Metadata' sheet "
        "contains any illegal characters and respects the regex expression"
    )

    def __init__(self, cdf_space_name, regex_expression, verbose=False):
        self.cdf_space_name = cdf_space_name
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid cdfSpaceName '{self.cdf_space_name}' stored in 'Metadata' sheet, "
            f"it must obey regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error102(NeatError):
    type_: str = "NamespaceNotValidURL"
    code: int = 102
    description: str = "namespace, which is in the 'Metadata' sheet, is not valid URL"
    example: str = "If we have 'authority:namespace' as namespace as it is not a valid URL this error will be raised"
    fix: str = (
        "Check if 'namespace' in the 'Metadata' sheet is properly "
        "constructed as valid URL containing only allowed characters"
    )

    def __init__(self, namespace, verbose=False):
        self.namespace = namespace

        self.message = f"Invalid namespace '{self.namespace}' stored in 'Metadata' sheet, it must be valid URL!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error103(NeatError):
    type_: str = "DataModelNameRegexViolation"
    code: int = 103
    description: str = "dataModelName, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If dataModelName is set to 'power grid data model', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if dataModelName in the 'Metadata' sheet contains any illegal "
        "characters and respects the regex expression"
    )

    def __init__(self, data_model_name, regex_expression, verbose=False):
        self.data_model_name = data_model_name
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid dataModelName '{self.data_model_name}' stored in 'Metadata' sheet, "
            f"it must obey regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error104(NeatError):
    type_: str = "VersionRegexViolation"
    code: int = 104
    description: str = "version, which is in the 'Metadata' sheet, does not respect defined regex expression"
    example: str = (
        "If version is set to '1.2.3 alpha4443', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if version in the 'Metadata' sheet contains any illegal characters and respects the regex expression"
    )

    def __init__(self, version, regex_expression, verbose=False):
        self.version = version
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid version '{self.version}' stored in 'Metadata' sheet, it must obey regex {self.regex_expression}!"
            f"\n\tFor further information visit https://thisisneat.io/errors/Error{self.code}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Warning1(NeatWarning):
    type_: str = "OWLGeneratedTransformationRulesHasErrors"
    code: int = 1
    description: str = (
        "This warning occurs when generating transformation rules from OWL ontology are invalid/incomplete."
    )
    example: str = ""
    fix: str = "Go through the generated report file and fix the warnings in generated Transformation Rules."

    def __init__(self, verbose=False):
        self.message = (
            "Transformation rules generated from OWL ontology are invalid!"
            " Consult report.txt for details on the errors and fix them before using the rules file."
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class Warning2(NeatWarning):
    type_: str = "OWLGeneratedTransformationRulesHasWarnings"
    code: int = 2
    description: str = (
        "This warning occurs when generating transformation rules from OWL ontology are invalid/incomplete."
    )
    example: str = ""
    fix: str = "Go through the generated report file and fix the warnings in generated Transformation Rules."

    def __init__(self, verbose=False):
        self.message = (
            "Transformation rules generated from OWL ontology raised warnings!"
            " Consult report.txt for details on warnings, and fix them prior using the rules file."
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class Warning30(NeatWarning):
    type_: str = "OntologyMultiTypeProperty"
    code: int = 30
    description: str = (
        "This warning occurs when a same property is define for two object"
        " where its expected value type is different in one case it acts as"
        " a node edge (i.e. object) in other case it acts as a node attribute"
        " (i.e. hold simple values such as strings)."
    )
    example: str = ""
    fix: str = "If a property takes different value types for different objects, simply define new property"

    def __init__(self, property_id: str = "", types: list[str] = [], verbose=False):
        self.message = (
            "It is bad practice to have multi type property! "
            f"Currently property '{property_id}' is defined as multi type property: {', '.join(types)}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning31(NeatWarning):
    type_: str = "OntologyMultiRangeProperty"
    code: int = 31
    description: str = (
        "This warning occurs when a property takes range of values" " which consists of union of multiple value types."
    )
    example: str = ""
    fix: str = "If a property takes different range of values, simply define new property for each range"

    def __init__(self, property_id: str = "", range_of_values: list[str] = [], verbose=False):
        self.message = (
            "Property should ideally have only single range of values. "
            f"Currently property '{property_id}' has multiple ranges: {', '.join(range_of_values)}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning32(NeatWarning):
    type_: str = "OntologyMultiDomainProperty"
    code: int = 32
    description: str = "This warning occurs when a property is reused/redefined for more than one classes."
    example: str = ""
    fix: str = (
        "No need to fix this, but make sure that property type is consistent"
        " across different classes and that ideally takes the same range of values"
    )

    def __init__(self, property_id: str = "", classes: list[str] = [], verbose=False):
        self.message = (
            "Property should ideally defined for single class. "
            f"Currently property '{property_id}' is defined for multiple classes: {', '.join(classes)}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning33(NeatWarning):
    type_: str = "OntologyMultiLabeledProperty"
    code: int = 33
    description: str = (
        "This warning occurs when a property is given multiple labels,"
        " typically if the same property is defined for different "
        "classes but different name is given."
    )
    example: str = ""
    fix: str = "This would be automatically fixes by taking the first name."

    def __init__(self, property_id: str = "", names: list[str] = [], verbose=False):
        self.message = (
            "Property should have single preferred label (human readable name)."
            f"Currently property '{property_id}' has multiple preferred labels: {', '.join(names)} !"
            f"Only the first name, i.e. '{names[0]}' will be considered!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning34(NeatWarning):
    type_: str = "OntologyMultiDefinitionProperty"
    code: int = 34
    description: str = (
        "This warning occurs when a property is given multiple human readable definitions,"
        " typically if the same property is defined for different "
        "classes and their usage differs from class to class."
    )
    example: str = ""
    fix: str = "This would be automatically fixes by concatenating all definitions."

    def __init__(self, property_id: str, verbose=False):
        self.message = (
            f"Multiple definitions (aka comments) of property '{property_id}' detected."
            " Definitions will be concatenated."
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning40(NeatWarning):
    type_: str = "FieldNotFoundInInstance"
    code: int = 40
    description: str = (
        "This warning occurs when a property, associated to the pydantic field, is not found in the instance."
        "The missing field will be removed, which might lead to failure of the pydantic model validation if"
        " the field/property is mandatory."
    )
    example: str = ""
    fix: str = (
        "If property/field is mandatory make sure that instances contain all mandatory fields."
        "Otherwise, no need to fix this warning."
    )

    def __init__(self, id_: str | URIRef = "", field_name: str = "", verbose=False):
        self.message = (
            f"Field {field_name} is missing in the instance {id_}."
            " If this field is mandatory, the validation of the pydantic model will fail!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning41(NeatWarning):
    type_: str = "FieldContainsMoreThanOneValue"
    code: int = 41
    description: str = (
        "This warning occurs when a property, associated to the pydantic field, contains"
        " more than one value (i.e. list of values), while it is defined as single value field."
        " As consequence, only the first value will be considered!"
    )
    example: str = ""
    fix: str = (
        "If a property takes more than one value, define it as list of values in TransformationRules."
        "To do this do not bound its `max_count` to 1, either leave it blank or set it to >1."
    )

    def __init__(self, field_name: str = "", no_of_values: int = None, verbose=False):
        self.message = (
            f"Field {field_name} is defined as single value property in TransformationRules,"
            f" but it contains {no_of_values} values!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Error60(NeatError):
    type_: str = "DataModelOrItsComponentsAlreadyExist"
    code: int = 60
    description: str = "This error is raised when attempting to create data model which already exist in DMS."
    example: str = ""
    fix: str = (
        "Remove existing data model and underlying views and/or containers, or bump "
        "version of data model and views and optionally delete containers."
    )

    def __init__(self, existing_data_model, existing_containers, existing_views, verbose=False):
        self.existing_data_model = existing_data_model
        self.existing_containers = existing_containers
        self.existing_views = existing_views

        self.message = "Aborting data model creation!"
        if self.existing_data_model:
            self.message += (
                f"\nData model {self.existing_data_model} already exists in DMS! Delete it first or bump its version! "
            )
        if self.existing_views:
            self.message += (
                f"\nViews {self.existing_views} already exist in DMS! Delete them first or bump their versions! "
            )
        if self.existing_containers:
            self.message += f"\nContainers {self.existing_containers} already exist in DMS! Delete them first! "

        self.message += (
            "\nTo remove existing data model and its components, use `self.remove_data_model(client)` method."
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Warning60(NeatWarning):
    type_: str = "ContainerPropertyTypeUnsupported"
    code: int = 60
    description: str = (
        "This warning occurs when a property type is not supported by the container."
        " Currently only `DatatypeProperty` and `ObjectProperty` are supported, which"
        " translate to `attribute` and `edge` respectively."
    )
    example: str = ""
    fix: str = "Contact NEAT support team."

    def __init__(self, property_id: str = "", property_type: str = "", verbose=False):
        self.message = (
            f"Property {property_id} has unsupported type {property_type}!"
            "Only DatatypeProperty and ObjectProperty are supported!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning61(NeatWarning):
    type_: str = "ViewPropertyTypeUnsupported"
    code: int = 61
    description: str = (
        "This warning occurs when a TransformationRule property translates to unsupported DMS view property."
        " Currently only attributes, edges 1-1 and edges 1-n are supported."
    )
    example: str = ""
    fix: str = "Contact NEAT support team."

    def __init__(self, property_id: str = "", verbose=False):
        self.message = (
            f"Property {property_id} translates to unsupported!"
            " Currently only attributes, edges 1-1 and edges 1-n are supported."
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning62(NeatWarning):
    type_: str = "ContainersAlreadyExist"
    code: int = 62
    description: str = "This warning occurs when attempting to create containers which already exist in DMS."
    example: str = ""
    fix: str = "Remove existing containers and try again."

    def __init__(self, container_ids: set[str] = set(), space: str = "", verbose=False):
        self.message = (
            f"Containers {container_ids} already exist in space {space}. "
            "Since update of containers can cause issues, "
            "remove them first prior data model creation!"
            "Aborting containers creation!"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning63(NeatWarning):
    type_: str = "ViewsAlreadyExist"
    code: int = 63
    description: str = "This warning occurs when attempting to create views which already exist in DMS."
    example: str = ""
    fix: str = "Remove existing views and try again or update version of data model."

    def __init__(self, views_ids: set[str] = set(), version: str = "", space: str = "", verbose=False):
        self.message = (
            f"Views {views_ids} version {version} already exist in space {space}. "
            "Since update of views raise issues, "
            "remove them first prior data model creation or update version of data model!"
            "Aborting views creation!"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning64(NeatWarning):
    type_: str = "DataModelAlreadyExist"
    code: int = 64
    description: str = "This warning occurs when attempting to create data model which already exist in DMS."
    example: str = ""
    fix: str = "Remove existing data model and try again or update its version."

    def __init__(self, data_model_id: str = "", version: str = "", space: str = "", verbose=False):
        self.message = (
            f"Data model {data_model_id} version {version} already exist in space {space}. "
            "Since update of data model can raise issues, "
            "remove it first or update its version!"
            "Aborting data model creation!"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning100(NeatWarning):
    type_: str = "NamespaceEndingFixed"
    code: int = 100
    description: str = "It is expected that namespace ends with '/' or '#'. If not, it will be fixed"
    example: str = "If namespace is set to http://purl.org/cognite, it will be converted to http://purl.org/cognite#"
    fix: str = "Make sure that namespace ends with '/' or '#'"

    def __init__(self, namespace, verbose=False):
        self.message = f"Namespace {namespace} ending fixed by adding '#' at its end!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class Warning101(NeatWarning):
    type_: str = "DataModelNameMissing"
    code: int = 101
    description: str = "In case when data model name is not provided in the 'Metadata' sheet, it will be set to prefix"
    example: str = ""
    fix: str = "Provide data model name to avoid this warning and defaulting to prefix"

    def __init__(self, prefix, verbose=False):
        self.message = f"Data model name not provided, defaulting to prefix {prefix}!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class Warning102(NeatWarning):
    type_: str = "VersionDotsConvertedToUnderscores"
    code: int = 102
    description: str = (
        "Version is expressed in classical form with dots major.minor.patch, "
        "while CDF accepts underscores major_minor_patch"
    )
    example: str = "If version is provided as 1.2.3, this will be converted to 1_2_3 to be accepted by CDF"
    fix: str = "Convert version to underscore notation major_minor_patch"

    def __init__(self, verbose=False):
        self.message = (
            "Data model version expressed with '.' which is illegal character for CDF. All '.' are converted to '_'!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


########################################
# Classes sheet Error Codes 200 - 199: #


class Error200(NeatError):
    type_: str = "ClassIDRegexViolation"
    code: int = 200
    description: str = (
        "Class ID, which is stored in the column 'Class' in the 'Classes' sheet, "
        "does not respect defined regex expression"
    )
    example: str = (
        "If class id is set to 'Class 1', while regex expression does not allow spaces,"
        " the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of class ids in 'Class' column in 'Classes' sheet and "
        "make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, class_id, regex_expression, verbose=False):
        self.class_id = class_id
        self.regex_expression = regex_expression

        self.message = (
            f"Class id '{self.class_id}' stored in 'Class' column in 'Classes' "
            f"sheet violates regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error201(NeatError):
    type_: str = "ClassIDMissing"
    code: int = 200
    description: str = (
        "Class ID, which is stored in the column 'Class' in the 'Classes' sheet,"
        " is either missing or did not satisfied regex expression"
    )
    example: str = ""
    fix: str = "Make sure that class id is provided and respects regex expression"

    def __init__(self, verbose=False):
        self.message = (
            "Class id is missing, it failed validation either because it has"
            " not been provided or because it did not respect regex expression!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Warning200(NeatWarning):
    type_: str = "ClassNameNotProvided"
    code: int = 200
    description: str = (
        "If class name is not provided in the 'Classes' sheet under 'name' column,"
        " it will be set to corresponding value from 'Class' column, thus class id"
    )
    example: str = ""
    fix: str = "If you want to have different class name than class id, provide it in the 'name' column"

    def __init__(self, class_id, verbose=False):
        self.message = f"Class id {class_id} set as Class name!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


###########################################
# Properties sheet Error Codes 300 - 399: #


class Error300(NeatError):
    type_: str = "ClassIDRegexViolation"
    code: int = 300
    description: str = (
        "Class ID, which is stored in the column 'Class' in the 'Properties' sheet, "
        "does not respect defined regex expression"
    )
    example: str = (
        "If class id is set to 'Class 1', while regex expression does not allow spaces,"
        " the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of class ids in 'Class' column in 'Properties' sheet "
        "and make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, class_id, regex_expression, verbose=False):
        self.class_id = class_id
        self.regex_expression = regex_expression

        self.message = (
            f"Class id '{self.class_id}' stored in 'Class' column in 'Properties' "
            f"sheet violates regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error301(NeatError):
    type_: str = "PropertyIDRegexViolation"
    code: int = 301
    description: str = (
        "Property ID, which is stored in the column 'Property' "
        "in the 'Properties' sheet, does not respect defined regex expression"
    )
    example: str = (
        "If property id is set to 'property 1', while regex expression does not allow spaces,"
        " the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of property ids in 'Property' column in 'Properties' sheet"
        " and make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, property_id, regex_expression, verbose=False):
        self.property_id = property_id
        self.regex_expression = regex_expression

        self.message = (
            f"Property id '{self.property_id}' stored in 'Property' "
            f"column in 'Properties' sheet violates regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error302(NeatError):
    type_: str = "ValueTypeIDRegexViolation"
    code: int = 302
    description: str = (
        "Value type, which is stored in the column 'Type' in the 'Properties' sheet, "
        "does not respect defined regex expression"
    )
    example: str = (
        "If value type is set to 'date time', while regex expression does not"
        " allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check definition of value types in 'Type' column in 'Properties' sheet"
        " and make sure to respect the regex expression by removing any illegal characters"
    )

    def __init__(self, value_type, regex_expression, verbose=False):
        self.value_type = value_type
        self.regex_expression = regex_expression

        self.message = (
            f"Value type '{self.value_type}' stored in 'Type' column in "
            f"'Properties' sheet violates regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error303(NeatError):
    type_: str = "MissingTypeValue"
    code: int = 302
    description: str = "Value type, which is stored in the column 'Type' in the 'Properties' sheet, is missing"
    example: str = "If value type is not set, this error will be raised"
    fix: str = "Make sure to define value type in 'Type' column in 'Properties' sheet"

    def __init__(self, verbose=False):
        self.message = "Value type, which is stored in the column 'Type' in the 'Properties' sheet, is missing"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error304(NeatError):
    type_: str = "PropertyIDMissing"
    code: int = 304
    description: str = (
        "Property ID, which is stored in the column 'Property' in the 'Properties' sheet,"
        " is either missing or did not satisfied regex expression"
    )
    example: str = ""
    fix: str = "Make sure that property id is provided and respects regex expression"

    def __init__(self, verbose=False):
        self.message = (
            "Property id is missing, validator for property id failed either"
            " due to lack of property id or due to not respecting regex expression!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error305(NeatError):
    type_: str = "RuleTypeProvidedButRuleMissing"
    code: int = 305
    description: str = (
        "This error occurs when transformation rule type is provided but actual transformation rule is missing"
    )
    example: str = ""
    fix: str = (
        "If you provide rule type you are must provide rule as well! "
        "Otherwise remove rule type if no transformation rule is needed"
    )

    def __init__(self, property_id, class_id, rule_type, verbose=False):
        self.message = (
            f"Rule type '{rule_type}' provided for property '{property_id}' "
            f"in class '{class_id}' but rule is not provided!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Warning300(NeatWarning):
    type_: str = "PropertyNameNotProvided"
    code: int = 300
    description: str = (
        "If property name is not provided in the 'Property' sheet under 'name' column,"
        " it will be set to corresponding value from 'Property' column, thus property id"
    )
    example: str = ""
    fix: str = "If you want to have different property name than property id, provide it in the 'name' column"

    def __init__(self, property_id, verbose=False):
        self.message = f"Property id {property_id} set as Property name!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning301(NeatWarning):
    type_: str = "MissingLabel"
    code: int = 301
    description: str = (
        "If property maps to CDF relationship, and it does not have label explicitly stated under 'Label' column"
        " in the 'Property' sheet under 'name' column,"
        " it will be set to corresponding value from 'Property' column, thus property id"
    )
    example: str = ""
    fix: str = (
        "If you want to have control over relationship labels make sure to define one"
        " in the 'Label' column in the 'Properties' sheet."
    )

    def __init__(self, property_id, verbose=False):
        self.message = f"Property id {property_id} set as CDF relationship label!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning302(NeatWarning):
    type_: str = "NoTransformationRules"
    code: int = 302
    description: str = (
        "This warning is raised if there are no transformation rules "
        "defined in the 'Transformation' sheet for given propertuy"
    )
    example: str = ""
    fix: str = "No fix is provided for this warning"

    # need to have default value set otherwise
    # warnings.warn(Warning302(property_id, class_id).message, category=Warning302)
    # will raise TypeError: __init__() missing 1 required positional argument: 'property_id'
    # not happy with this solution but it works
    def __init__(self, property_id: str = "", class_id: str = "", verbose=False):
        self.message = f"There is no transformation rule configured for class '{class_id}' property '{property_id}'!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


###############################################
#  Prefixes  Error Codes 400 - 499: #


class Error400(NeatError):
    type_: str = "PrefixesRegexViolation"
    code: int = 400
    description: str = "Prefix(es), which are in the 'Prefixes' sheet, do(es) not respect defined regex expression"
    example: str = (
        "If prefix is set to 'power grid', while regex expression does not "
        "allow spaces, the expression will be violated thus raising this error"
    )
    fix: str = (
        "Check if prefixes in the 'Prefixes' sheet contains any illegal characters and respects the regex expression"
    )

    def __init__(self, prefixes, regex_expression, verbose=False):
        self.prefixes = prefixes
        self.regex_expression = regex_expression

        self.message = (
            f"Invalid prefix(es) {', '.join(self.prefixes)} stored in the 'Prefixes' sheet, "
            f"it/they must obey regex {self.regex_expression}!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error401(NeatError):
    type_: str = "NamespaceNotValidURL"
    code: int = 401
    description: str = "namespace(es), which are/is in the 'Prefixes' sheet, are/is not valid URLs"
    example: str = "If we have 'authority:namespace' as namespace as it is not a valid URL this error will be raised"
    fix: str = (
        "Check if 'namespaces' in the 'Prefixes' sheet are properly "
        "constructed as valid URLs containing only allowed characters"
    )

    def __init__(self, namespaces, verbose=False):
        self.namespaces = namespaces

        self.message = (
            f"Invalid namespace(es) {', '.join(self.namespaces)} stored in the 'Prefixes' sheet, "
            f"it/they must be valid URLs!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


###############################################
# Instances Error Codes 500 - 599: #


class Warning500(NeatWarning):
    type_: str = "MissingDataModelPrefixOrNamespace"
    code: int = 500
    description: str = "Either prefix or namespace or both are missing in the 'Metadata' sheet"
    example: str = ""
    fix: str = "Add missing prefix and/or namespace in the 'Metadata' sheet"

    def __init__(self, verbose=False):
        self.message = (
            "Instances sheet is present but prefix and/or namespace are missing in 'Metadata' sheet."
            "Instances sheet will not be processed!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


###############################################
# Transformation Rules Error Codes 600 - 699: #


class Warning600(NeatWarning):
    type_: str = "EntityIDNotDMSCompliant"
    code: int = 600
    description: str = "Warning raise when entity id being class, property or value type is not DMS compliant"
    example: str = ""
    fix: str = (
        "DMS ready means that entity id must only use following"
        " characters [a-zA-Z0-9_], where it can only start with letter!"
    )

    # See Warning302 for explanation why default values are set
    def __init__(self, entity_type: str = "", entity_id: str = "", loc: str = "", verbose=False):
        self.message = (
            f"'{entity_id}' {entity_type.lower()}"
            " use character(s) outside of range of allowed characters [a-zA-Z0-9_] or "
            f"it starts with non-letter character! {loc}"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"


class Warning601(NeatWarning):
    type_: str = "PropertyRedefined"
    code: int = 600
    description: str = "Warning raise when same property is defined multiple times for same class"
    example: str = ""
    fix: str = "Have only single definition of a perticular property for a class"

    # See Warning302 for explanation why default values are set
    def __init__(self, property_id: str = "", class_id: str = "", loc: str = "", verbose=False):
        self.message = f"Not DMS compliant! Property '{property_id}' for class '{class_id}' redefined! {loc}"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
            # hint on a specific web docs page


class Error600(NeatError):
    """Property defined for a class that has not been defined in the 'Classes' sheet"""

    type_: str = "PropertyDefinedForUndefinedClass"
    code: int = 600
    description: str = "Property defined for a class that has not been defined in the 'Classes' sheet"
    example: str = (
        "If property 'someProperty' is defined for class 'Class 1' in the 'Properties' sheet, "
        "while 'Class 1' has not been defined in the 'Classes' sheet,"
        " this error will be raised"
    )
    fix: str = (
        "Make sure to define all classes in the 'Classes' sheet before defining properties for them"
        " in the 'Properties' sheet"
    )

    def __init__(self, property_id, class_id, verbose=False):
        self.property_id = property_id
        self.class_id = class_id

        self.message = (
            f"Class <{self.class_id}> to which property {self.property_id}> is being defined"
            " is not define in the 'Classes' sheet!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error601(NeatError):
    type_: str = "MetadataSheetMissingOrFailedValidation"
    code: int = 601
    description: str = "Metadata sheet is missing or it failed validation for one or more fields"
    example: str = ""
    fix: str = "Make sure to define compliant Metadata sheet before proceeding"

    def __init__(self, verbose=False):
        self.message = "Metadata sheet is missing or it failed validation for one or more fields!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error602(NeatError):
    type_: str = "FiledInMetadataSheetMissingOrFailedValidation"
    code: int = 602
    description: str = "One of the expected fields in Metadata sheet is missing or it failed validation"
    example: str = ""
    fix: str = "Make sure to define compliant field in Metadata sheet before proceeding"

    def __init__(self, missing_field: str, verbose=False):
        self.message = f"Field {missing_field} is missing in the 'Metadata' sheet or it failed validation!"
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error603(NeatError):
    type_: str = "ValueTypeNotDefinedAsClass"
    code: int = 603
    description: str = (
        "Expected value type, which is stored in the column 'Type' in the 'Properties'"
        " sheet, is not defined in the 'Classes' sheet. "
        "This error occurs when property is defined as an edge between two classes, of which one is not defined"
    )
    example: str = (
        "We have 'Class1' which has property 'edgeClass1Class2' linking it to 'Class2', thus"
        "expected value of 'edgeClass1Class2' is 'Class2'. However, 'Classes' sheet only contains"
        " 'Class1', while 'Class2' is not defined. Under this given circumstance, this error will be raised!"
    )

    fix: str = (
        "Make sure to define all of the classes in the 'Classes' sheet before defining "
        "properties that expect them as value types"
    )

    def __init__(self, class_id: str, property_id: str, expected_value_type: str, verbose=False):
        self.message = (
            f"Property {property_id} defined for class {class_id} has"
            f" value type {expected_value_type} which is not defined as a class in the 'Classes' sheet!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class Error604(NeatError):
    type_: str = "UndefinedObjectsAsExpectedValueTypes"
    code: int = 604
    description: str = (
        "Expected value types, which are stored in the column 'Type' in the 'Properties'"
        " sheet, are classes that exist in the 'Classes' sheet but for which no properties are defined "
        "in the 'Properties' sheet. "
    )
    example: str = (
        "We have 'Class1' which has property 'edgeClass1Class2' linking it to 'Class2', thus"
        "expected value of 'edgeClass1Class2' is 'Class2'. "
        "Both 'Class1' and 'Class2' are defined in the 'Classes' sheet"
        "However, only 'Class1' has properties defined in the 'Properties' sheet, making 'Class2' an undefined object"
        " leading to this error being raised!"
    )

    fix: str = (
        "Make sure to define properties for classes from 'Classes' "
        "sheet before defining properties that expect them as value types"
    )

    def __init__(self, undefined_objects: list[str], verbose=False):
        self.message = (
            f"Following classes {', '.join(undefined_objects)} defined as classes in the 'Classes' sheet"
            f" have no properties defined in the 'Properties' sheet or their validation as objects failed!"
        )
        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)



import logging
from pathlib import Path
import warnings

import pandas as pd
import numpy as np
from openpyxl import Workbook, load_workbook
from rdflib import RDF, XSD, Literal, Namespace

from cognite.neat.rules.analysis import to_class_property_pairs, get_defined_classes
from cognite.neat.rules.models import TransformationRules


def extract_graph_from_sheet(
    filepath: Path,
    transformation_rule: TransformationRules,
    separator: str = ",",
    namespace: str = None,
) -> list[tuple]:
    """Converts a graph capturing sheet to rdf triples

    Parameters
    ----------
    filepath : Path
        Path to the graph capturing sheet
    graph_capturing_sheet : dict[str, pd.DataFrame]
        Graph capturing sheet
    transformation_rule : TransformationRules
        Transformation rules
    separator : str, optional
        Multi value separator at cell level, by default ","
    namespace : str, optional
        In case of a custom namespace, by default None meaning the namespace is taken from the transformation rules
    """

    graph_capturing_sheet = read_graph_excel_file_to_table_by_name(filepath)

    return sheet2triples(graph_capturing_sheet, transformation_rule, separator, namespace)


def sheet2triples(
    graph_capturing_sheet: dict[str, pd.DataFrame],
    transformation_rule: TransformationRules,
    separator: str = ",",
    namespace: str = None,
) -> list[tuple]:
    """Converts a graph capturing sheet to rdf triples

    Parameters
    ----------
    graph_capturing_sheet : dict[str, pd.DataFrame]
        Graph capturing sheet
    transformation_rule : TransformationRules
        Transformation rules
    separator : str, optional
        Multi value separator at cell level, by default ","
    namespace : str, optional
        In case of a custom namespace, by default None meaning the namespace is taken from the transformation rules
    """
    # Validation that everything is in order before proceeding
    validate_if_graph_capturing_sheet_empty(graph_capturing_sheet)
    validate_rules_graph_pair(graph_capturing_sheet, transformation_rule)

    # get class property pairs
    class_property_pairs = to_class_property_pairs(transformation_rule)

    # namespace selection
    if namespace is None:
        instance_namespace = transformation_rule.metadata.namespace
    else:
        instance_namespace = Namespace(namespace)

    model_namespace = Namespace(transformation_rule.metadata.namespace)

    # Now create empty graph
    triples = []

    # Add triples from the capturing sheet to the graph by iterating over the capturing sheet
    # iterate over sheets
    for sheet_name, df in graph_capturing_sheet.items():
        # iterate over sheet rows
        for _, row in df.iterrows():
            if row.identifier is None:
                msg = f"Missing identifier in sheet {sheet_name} at row {row.name}! Skipping..."
                logging.warning(msg)
                warnings.warn(
                    msg,
                    stacklevel=2,
                )
                continue

            # iterate over sheet rows properties
            for property_, value in row.to_dict().items():
                # Setting RDF type of the instance
                if property_ == "identifier":
                    triples.append(
                        (
                            instance_namespace[row.identifier],
                            RDF.type,
                            model_namespace[sheet_name],
                        )
                    )

                # Adding object properties
                elif class_property_pairs[sheet_name][property_].property_type == "ObjectProperty" and value:
                    triples.extend(
                        (
                            instance_namespace[row.identifier],
                            model_namespace[property_],
                            instance_namespace[v.strip()],
                        )
                        for v in value.split(separator)
                    )

                # Adding data properties
                # TODO: Add support for datatype
                elif value:
                    triples.append(
                        (
                            instance_namespace[row.identifier],
                            model_namespace[property_],
                            Literal(
                                value, datatype=XSD[class_property_pairs[sheet_name][property_].expected_value_type]
                            ),
                        )
                    )
    return triples


def validate_if_graph_capturing_sheet_empty(graph_capturing_sheet: dict[str, pd.DataFrame]):
    """Validate if the graph capturing sheet is empty

    Parameters
    ----------
    graph_capturing_sheet : dict[str, pd.DataFrame]
        Graph capturing sheet
    """
    if all(df.empty for df in graph_capturing_sheet.values()):
        msg = "Graph capturing sheet is empty! Aborting!"
        logging.error(msg)
        raise ValueError(msg)


def validate_rules_graph_pair(graph_capturing_sheet: dict[str, pd.DataFrame], transformation_rule: TransformationRules):
    """Validate if the graph capturing sheet is based on the transformation rules

    Parameters
    ----------
    graph_capturing_sheet : dict[str, pd.DataFrame]
        Graph capturing sheet
    transformation_rule : TransformationRules
        Transformation rules
    """
    intersection = set(graph_capturing_sheet.keys()).intersection(set(get_defined_classes(transformation_rule)))

    if not intersection:
        msg = "Graph capturing sheet is not based on transformation rules! Aborting!"
        logging.error(msg)
        raise ValueError(msg)

    elif len(intersection) == len(graph_capturing_sheet.keys()):
        logging.info("All classes in the graph capturing sheet are defined in the transformation rules!")

    elif len(intersection) < len(graph_capturing_sheet.keys()):
        msg = "Graph capturing sheet contains classes that are not defined in the transformation rules! Proceeding..."
        logging.warning(msg)
        warnings.warn(
            msg,
            stacklevel=2,
        )

    elif len(intersection) < len(get_defined_classes(transformation_rule)):
        msg = "Transformation rules contain classes that are not present in the graph capturing sheet! Proceeding..."
        logging.warning(msg)
        warnings.warn(
            msg,
            stacklevel=2,
        )


def read_graph_excel_file_to_table_by_name(filepath: Path) -> dict[str, pd.DataFrame]:
    workbook: Workbook = load_workbook(filepath)

    parsed_sheets = {
        sheetname: pd.read_excel(
            filepath,
            sheet_name=sheetname,
            header=0,
        )
        for sheetname in workbook.sheetnames
    }

    for sheetname, df in parsed_sheets.items():
        if "identifier" in df.columns:
            parsed_sheets[sheetname] = df.drop(df[df.identifier == 0].index)
            parsed_sheets[sheetname] = df.replace({np.nan: None})
        else:
            logging.error(f"Sheet {sheetname} does not have an identifier column")
            raise ValueError(f"Sheet {sheetname} does not have an identifier column")

    return parsed_sheets



import re
from typing import Literal, overload
import warnings

from cognite.neat.rules import _exceptions
from cognite.neat.rules.models import TransformationRules, data_model_name_compliance_regex


@overload
def are_entity_names_dms_compliant(
    transformation_rules: TransformationRules, return_report: Literal[True]
) -> tuple[bool, list[dict]]:
    ...


@overload
def are_entity_names_dms_compliant(
    transformation_rules: TransformationRules, return_report: Literal[False] = False
) -> bool:
    ...


def are_entity_names_dms_compliant(
    transformation_rules: TransformationRules, return_report: bool = False
) -> bool | tuple[bool, list[dict]]:
    """Check if data model definitions are valid."""

    flag: bool = True
    with warnings.catch_warnings(record=True) as validation_warnings:
        for class_ in transformation_rules.classes.values():
            if not re.match(data_model_name_compliance_regex, class_.class_id):
                warnings.warn(
                    _exceptions.Warning600("Class", class_.class_id, f"[Classes/Class/{class_.class_id}]").message,
                    category=_exceptions.Warning600,
                    stacklevel=2,
                )
                flag = False

        for row, property_ in transformation_rules.properties.items():
            if not re.match(data_model_name_compliance_regex, property_.class_id):
                warnings.warn(
                    _exceptions.Warning600("Class", property_.class_id, f"[Properties/Class/{row}]").message,
                    category=_exceptions.Warning600,
                    stacklevel=2,
                )
                flag = False
            if not re.match(data_model_name_compliance_regex, property_.property_id):
                warnings.warn(
                    _exceptions.Warning600("Property", property_.property_id, f"[Properties/Property/{row}]").message,
                    category=_exceptions.Warning600,
                    stacklevel=2,
                )
                flag = False
            if not re.match(data_model_name_compliance_regex, property_.expected_value_type):
                warnings.warn(
                    _exceptions.Warning600(
                        "Value type", property_.expected_value_type, f"[Properties/Type/{row}]"
                    ).message,
                    category=_exceptions.Warning600,
                    stacklevel=2,
                )
                flag = False

    if return_report:
        return flag, _exceptions.wrangle_warnings(validation_warnings)
    else:
        return flag


@overload
def are_properties_redefined(
    transformation_rules: TransformationRules, return_report: Literal[True]
) -> tuple[bool, list[dict]]:
    ...


@overload
def are_properties_redefined(transformation_rules: TransformationRules, return_report: Literal[False] = False) -> bool:
    ...


def are_properties_redefined(
    transformation_rules: TransformationRules, return_report: bool = False
) -> bool | tuple[bool, list[dict]]:
    flag: bool = False
    with warnings.catch_warnings(record=True) as validation_warnings:
        analyzed_properties = {}
        for property_ in transformation_rules.properties.values():
            if property_.property_id not in analyzed_properties:
                analyzed_properties[property_.property_id] = [property_.class_id]
            elif property_.class_id in analyzed_properties[property_.property_id]:
                flag = True
                warnings.warn(
                    _exceptions.Warning601(property_.class_id, property_.property_id).message,
                    category=_exceptions.Warning600,
                    stacklevel=2,
                )

            else:
                analyzed_properties[property_.property_id].append(property_.class_id)

    if return_report:
        return flag, _exceptions.wrangle_warnings(validation_warnings)
    else:
        return flag


def property_ids_camel_case_compliant(transformation_rules) -> bool | tuple[bool, list[dict]]:
    ...


def class_id_pascal_case_compliant(transformation_rules) -> bool | tuple[bool, list[dict]]:
    ...



from cognite.neat.utils.exceptions import NeatError


class InvalidStepOutputException(NeatError):
    type_ = "invalid_step_output"
    code = 400
    description = "The step output is invalid."
    example = "The step output must be a dictionary."



from typing import Iterable

from cognite.client import CogniteClient
from prometheus_client import REGISTRY, Counter, Gauge, Metric


class NeatMetricsCollector:
    def __init__(self, name: str, cdf_client: CogniteClient = None) -> None:
        self.name = name
        self.metrics: dict[str, Gauge | Counter] = {}

    def register_metric(
        self, metric_name: str, metric_description: str = "", m_type: str = "gauge", metric_labels: list[str] = None
    ) -> Gauge | Counter:
        """Register metric in prometheus"""
        metric_labels = [] if metric_labels is None else metric_labels

        metric_name = f"neat_workflow_{self.name}_{metric_name}"
        if metric_name in REGISTRY._names_to_collectors:
            self.metrics[metric_name] = REGISTRY._names_to_collectors[metric_name]
            return self.metrics[metric_name]

        metric = None
        if m_type == "gauge":
            metric = Gauge(metric_name, metric_description, metric_labels)
        elif m_type == "counter":
            metric = Counter(metric_name, metric_description, metric_labels)

        if metric:
            self.metrics[metric_name] = metric
            return metric

    def get(self, metric_name: str, labels: dict[str, str] = None) -> Gauge | Counter:
        """Return metric by name"""
        labels = {} if labels is None else labels
        metric_name = f"neat_workflow_{self.name}_{metric_name}"
        return self.metrics.get(metric_name).labels(**labels)

    def report_metric_value(
        self, metric_name: str, metric_description: str = "", m_type: str = "gauge", labels: dict[str, str] = None
    ) -> Gauge | Counter:
        metric = self.register_metric(metric_name, metric_description, m_type, [k for k, v in labels.items()])
        return metric.labels(**labels)

    def collect(self) -> Iterable[Metric]:
        pass

    def report_to_cdf():
        pass



import logging
import uuid
from pathlib import Path

from openpyxl import Workbook
from openpyxl.styles import Alignment, Border, Font, NamedStyle, PatternFill, Side
from openpyxl.utils.cell import get_column_letter
from openpyxl.worksheet.datavalidation import DataValidation

from cognite.neat.rules.analysis import to_class_property_pairs
from cognite.neat.rules.exporter.rules2rules import to_dms_name
from cognite.neat.rules.models import TransformationRules


def rules2graph_capturing_sheet(
    transformation_rules: TransformationRules,
    file_path: Path,
    no_rows: int = 1000,
    auto_identifier_type: str = "index-based",
    add_drop_down_list: bool = True,
):
    """Converts a TransformationRules object to a graph capturing sheet

    Parameters
    ----------
    transformation_rules : TransformationRules
        The TransformationRules object to convert to the graph capturing sheet
    file_path : Path
        File path to save the sheet to
    auto_identifier_type : str, optional
        Type of automatic identifier, by default "index" based
    no_rows : int, optional
        Number of rows for processing, by default 1000
    add_drop_down_list : bool, optional
        Add drop down selection for columns that contain linking properties, by default True

    Notes
    -----
    no_rows should be set to the maximum expected number of instances of any of the classes.
    By default, it is set to 1000, increase it accordingly if you have more instances.

    """

    workbook = Workbook()

    # Remove default sheet named "Sheet"
    workbook.remove(workbook["Sheet"])

    for class_, properties in to_class_property_pairs(transformation_rules).items():
        workbook.create_sheet(title=class_)

        # Add header rows
        workbook[class_].append(["identifier"] + list(properties.keys()))

        if auto_identifier_type and auto_identifier_type == "index-based":  # default, easy to read
            logging.debug(f"Configuring index-based automatic identifiers for sheet {class_}")
            _add_index_identifiers(workbook, class_, no_rows)
        elif auto_identifier_type and auto_identifier_type == "uuid-based":
            logging.debug(f"Configuring UUID-based automatic identifiers for sheet {class_}")
            _add_uuid_identifiers(workbook, class_, no_rows)
        else:
            logging.debug(f"No automatic identifier set for sheet {class_}!")

        for i, property_ in enumerate(properties.values()):
            if property_.property_type == "ObjectProperty" and add_drop_down_list:
                _add_drop_down_list(
                    workbook,
                    class_,
                    get_column_letter(i + 2),
                    no_rows,
                    property_.expected_value_type,
                    "A",
                )

    _adjust_column_width(workbook)
    _set_header_style(workbook)

    logging.info(f"Graph capturing sheet generated and stored at {file_path}!")
    workbook.save(file_path)
    workbook.close()


def _add_index_identifiers(workbook: Workbook, sheet: str, no_rows: int):
    """Adds index-based auto identifier to a sheet identifier column"""
    for i in range(no_rows):
        prefix = to_dms_name(sheet, "class", True)
        workbook[sheet][f"A{i+2}"] = f'=IF(ISBLANK(B{i+2}), "","{prefix}-{i+1}")'


def _add_uuid_identifiers(workbook: Workbook, sheet: str, no_rows: int):
    """Adds UUID-based auto identifier to a sheet identifier column"""
    for i in range(no_rows):
        prefix = to_dms_name(sheet, "class", True)
        workbook[sheet][f"A{i+2}"] = f'=IF(ISBLANK(B{i+2}), "","{prefix}-{uuid.uuid4()}")'


def _add_drop_down_list(workbook: Workbook, sheet: str, column: str, no_rows: int, value_sheet: str, value_column: str):
    """Adds a drop down list to a column"""
    drop_down_list = DataValidation(type="list", formula1=f"={value_sheet}!{value_column}$2:{value_column}${no_rows}")

    workbook[sheet].add_data_validation(drop_down_list)

    for i in range(no_rows):
        drop_down_list.add(workbook[sheet][f"{column}{i+2}"])


def _adjust_column_width(workbook: Workbook):
    """Adjusts the column width based on the content"""
    for sheet in workbook.sheetnames:
        for column in workbook[sheet].columns:
            if column[0].value:
                adjusted_width = (len(str(column[0].value)) + 5) * 1.2
                workbook[sheet].column_dimensions[column[0].column_letter].width = adjusted_width


def _set_header_style(workbook: Workbook):
    """Sets the header style for all sheets in the workbook"""
    style = NamedStyle(name="header style")
    style.font = Font(bold=True, size=16)
    side = Side(style="thin", color="000000")
    style.border = Border(left=side, right=side, top=side, bottom=side)
    workbook.add_named_style(style)

    for sheet in workbook.sheetnames:
        for column in workbook[sheet].columns:
            workbook[sheet][f"{column[0].column_letter}1"].style = style
            if f"{column[0].column_letter}1" == "A1":
                workbook[sheet][f"{column[0].column_letter}1"].fill = PatternFill("solid", start_color="2FB5F2")
            else:
                workbook[sheet][f"{column[0].column_letter}1"].fill = PatternFill("solid", start_color="FFB202")
            workbook[sheet][f"{column[0].column_letter}1"].alignment = Alignment(horizontal="center", vertical="center")



from cognite.neat.graph.loaders.core.labels import upload_labels
from .core.rdf_to_assets import rdf2assets, upload_assets, categorize_assets
from .core.rdf_to_relationships import rdf2relationships, categorize_relationships, upload_relationships

__all__ = [
    "rdf2relationships",
    "categorize_assets",
    "upload_assets",
    "rdf2assets",
    "categorize_relationships",
    "upload_relationships",
    "upload_labels",
]



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles
import pkg_resources


from cognite import neat
from cognite.neat.app.api.configuration import UI_PATH, neat_app
from cognite.neat.app.api.context_manager import lifespan
from cognite.neat.app.api.asgi.metrics import prometheus_app
from cognite.neat.app.api.routers import configuration, crud, metrics, workflows, rules, data_exploration


app = FastAPI(title="Neat", lifespan=lifespan)


origins = [
    "http://localhost:8000",
    "http://localhost:3000",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Mount ASGI apps
app.mount("/metrics", prometheus_app)
app.mount("/static", StaticFiles(directory=UI_PATH), name="static")
app.mount("/data", StaticFiles(directory=neat_app.config.data_store_path), name="data")


# Mount routers
app.include_router(configuration.router)
app.include_router(metrics.router)
app.include_router(workflows.router)
app.include_router(rules.router)
app.include_router(crud.router)
app.include_router(data_exploration.router)


# General routes
@app.get("/")
def read_root():
    return RedirectResponse("/static/index.html")


@app.get("/api/about")
def get_about():
    installed_packages = pkg_resources.working_set
    installed_packages_list = sorted([f"{i.key}=={i.version}" for i in installed_packages])
    return {"version": neat.__version__, "packages": installed_packages_list}



from collections import defaultdict
import warnings
import pandas as pd

from cognite.neat.rules.models import Property, TransformationRules
from cognite.neat.rules.to_rdf_path import RuleType


def get_defined_classes(transformation_rules: TransformationRules) -> set[str]:
    """Returns classes that have been defined in the data model."""
    return {property.class_id for property in transformation_rules.properties.values()}


def get_classes_with_properties(transformation_rules: TransformationRules) -> dict[str, list[Property]]:
    """Returns classes that have been defined in the data model."""
    # TODO: Do not particularly like method name, find something more suitable
    class_property_pairs = {}

    for property_ in transformation_rules.properties.values():
        class_ = property_.class_id
        if class_ in class_property_pairs:
            class_property_pairs[class_] += [property_]
        else:
            class_property_pairs[class_] = [property_]

    return class_property_pairs


def to_class_property_pairs(
    transformation_rules: TransformationRules, only_rdfpath: bool = False
) -> dict[str, dict[str, Property]]:
    """Returns a dictionary of classes with a dictionary of properties associated with them.

    Parameters
    ----------
    transformation_rules : TransformationRules
        Instance of TransformationRules holding the data model
    only_rdfpath : bool, optional
        To consider only properties which have rule `rdfpath` set, by default False

    Returns
    -------
    dict[str, dict[str, Property]]
        Dictionary of classes with a dictionary of properties associated with them.

    Notes
    -----
    If only_rdfpath is True, only properties with RuleType.rdfpath will be returned as
    a part of the dictionary of properties related to a class. Otherwise, all properties
    will be returned.
    """

    class_property_pairs = {}

    for class_, properties in get_classes_with_properties(transformation_rules).items():
        processed_properties = {}
        for property_ in properties:
            if property_.property_id in processed_properties:
                # TODO: use appropriate Warning class from _exceptions.py
                # if missing make one !
                warnings.warn(
                    "Property has been defined more than once! Only first definition will be considered.",
                    stacklevel=2,
                )
                continue

            if (only_rdfpath and property_.rule_type == RuleType.rdfpath) or not only_rdfpath:
                processed_properties[property_.property_id] = property_
        class_property_pairs[class_] = processed_properties

    return class_property_pairs


def get_class_linkage(transformation_rules: TransformationRules) -> pd.DataFrame:
    """Returns a dataframe with the class linkage of the data model."""

    class_linkage = pd.DataFrame(columns=["source_class", "target_class", "connecting_property", "max_occurrence"])
    for property_ in transformation_rules.properties.values():
        if property_.property_type == "ObjectProperty":
            new_row = pd.Series(
                {
                    "source_class": property_.class_id,
                    "target_class": property_.expected_value_type,
                    "connecting_property": property_.property_id,
                    "max_occurrence": property_.max_count,
                }
            )
            class_linkage = pd.concat([class_linkage, new_row.to_frame().T], ignore_index=True)

    class_linkage.drop_duplicates(inplace=True)

    return class_linkage


def get_connected_classes(transformation_rules: TransformationRules) -> set[str]:
    """Return a set of classes that are connected to other classes."""
    class_linkage = get_class_linkage(transformation_rules)
    return set(class_linkage.source_class.values).union(set(class_linkage.target_class.values))


def get_disconnected_classes(transformation_rules: TransformationRules):
    """Return a set of classes that are disconnected (i.e. isolated) from other classes."""
    return get_defined_classes(transformation_rules) - get_connected_classes(transformation_rules)


def get_symmetric_pairs(transformation_rules: TransformationRules) -> set[tuple]:
    """Returns a list of pairs of symmetrically linked classes."""
    # TODO: Find better name for this method
    sym_pairs = set()

    class_linkage = get_class_linkage(transformation_rules)
    if class_linkage.empty:
        return sym_pairs

    for _, row in class_linkage.iterrows():
        source = row.source_class
        target = row.target_class
        target_targets = class_linkage[class_linkage.source_class == target].target_class.values
        if source in target_targets and (source, target) not in sym_pairs:
            sym_pairs.add((source, target))
    return sym_pairs


def get_entity_ids(transformation_rules: TransformationRules) -> set[str]:
    return set(transformation_rules.classes.keys()).union(
        {property_.property_id for property_ in transformation_rules.properties.values()}
    )

    # Methods below could as well easily go to analysis.py


def to_property_dict(rules: TransformationRules) -> dict[str, list[Property]]:
    """Convert list of properties to a dictionary of lists of properties with property_id as key."""
    property_: dict[str, list[Property]] = defaultdict(list)

    for prop in rules.properties.values():
        if not (prop.property_id and prop.property_name == "*"):
            property_[prop.property_id].append(prop)

    return property_


def get_asset_related_properties(properties: list[Property]) -> list[Property]:
    return [prop for prop in properties if "Asset" in prop.cdf_resource_type]


def define_class_asset_mapping(transformation_rules: TransformationRules, class_: str) -> dict[str, list[str]]:
    mapping_dict: dict[str, list[str]] = {}

    class_properties = to_class_property_pairs(transformation_rules, only_rdfpath=True)[class_]

    for asset_property in get_asset_related_properties(list(class_properties.values())):
        for resource_type_property in asset_property.resource_type_property:
            if resource_type_property not in mapping_dict:
                mapping_dict[resource_type_property] = [asset_property.property_id]
            else:
                mapping_dict[resource_type_property] += [asset_property.property_id]

    return mapping_dict


def define_class_relationship_mapping(transformation_rules: TransformationRules, class_: str) -> dict[str, list[str]]:
    ...



import hashlib
import logging
import time
from collections import OrderedDict
from datetime import datetime, timezone
from functools import wraps
from typing_extensions import TypeAlias
import pandas as pd
from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import CredentialProvider, OAuthClientCredentials, OAuthInteractive
from cognite.client.exceptions import CogniteDuplicatedError, CogniteReadTimeout
from rdflib import Literal
from rdflib.term import URIRef

from cognite.neat.graph.stores import NeatGraphStore
from cognite.neat.utils.cdf import InteractiveCogniteClient, ServiceCogniteClient, CogniteClientConfig

Triple: TypeAlias = tuple[URIRef, URIRef, Literal | URIRef]


def get_cognite_client_from_config(config: ServiceCogniteClient) -> CogniteClient:
    credentials = OAuthClientCredentials(
        token_url=config.token_url, client_id=config.client_id, client_secret=config.client_secret, scopes=config.scopes
    )

    return _get_cognite_client(config, credentials)


def get_cognite_client_interactive(config: InteractiveCogniteClient) -> CogniteClient:
    credentials = OAuthInteractive(
        authority_url=config.authority_url,
        client_id=config.client_id,
        scopes=config.scopes,
        redirect_port=config.redirect_port,
    )
    return _get_cognite_client(config, credentials)


def _get_cognite_client(config: CogniteClientConfig, credentials: CredentialProvider) -> CogniteClient:
    logging.info(f"Creating CogniteClient with parameters : {config}")
    return CogniteClient(
        ClientConfig(
            client_name=config.client_name,
            base_url=config.base_url,
            project=config.project,
            credentials=credentials,
            timeout=config.timeout,
            max_workers=config.max_workers,
            debug=False,
        )
    )


def add_triples(graph_store: NeatGraphStore, triples: list[Triple], batch_size: int = 10000):
    """Adds triples to the graph store in batches.

    Parameters
    ----------
    graph_store : NeatGraphStore
        Instance of NeatGraphStore
    triples : list[tuple]
        list of triples to be added to the graph store
    batch_size : int, optional
        Batch size of triples per commit, by default 10000
    """

    commit_counter = 0
    logging.info(f"Committing total of {len(triples)} triples to knowledge graph!")
    total_number_of_triples = len(triples)
    number_of_uploaded_triples = 0

    def check_commit(force_commit: bool = False):
        """Commit nodes to the graph if batch counter is reached or if force_commit is True"""
        nonlocal commit_counter
        nonlocal number_of_uploaded_triples
        if force_commit:
            number_of_uploaded_triples += commit_counter
            graph_store.graph.commit()
            logging.info(f"Committed {number_of_uploaded_triples} of {total_number_of_triples} triples")
            return
        commit_counter += 1
        if commit_counter >= batch_size:
            number_of_uploaded_triples += commit_counter
            graph_store.graph.commit()
            logging.info(f"Committed {number_of_uploaded_triples} of {total_number_of_triples} triples")
            commit_counter = 0

    for triple in triples:
        graph_store.graph.add(triple)
        check_commit()

    check_commit(force_commit=True)


def remove_namespace(URI: URIRef | str, special_separator: str = "#_") -> str:
    """Removes namespace from URI

    Parameters
    ----------
    URI : URIRef | str
        URI of an entity
    special_separator : str
        Special separator to use instead of # or / if present in URI
        Set by default to "#_" which covers special client use case

    Returns
    -------
    str
        Entity id without namespace
    """

    return URI.split(special_separator if special_separator in URI else ("#" if "#" in URI else "/"))[-1]


def get_namespace(URI: URIRef, special_separator: str = "#_") -> str:
    """Removes namespace from URI

    Parameters
    ----------
    URI : URIRef
        URI of an entity
    special_separator : str
        Special separator to use instead of # or / if present in URI
        Set by default to "#_" which covers special client use case

    Returns
    -------
    str
        Entity id without namespace
    """
    if special_separator in URI:
        return URI.split(special_separator)[0] + special_separator
    elif "#" in URI:
        return URI.split("#")[0] + "#"
    else:
        return "/".join(URI.split("/")[:-1]) + "/"


def _traverse(hierarchy: dict, graph: dict, names: list[str]) -> dict:
    """traverse the graph and return the hierarchy"""
    for name in names:
        hierarchy[name] = _traverse({}, graph, graph[name])
    return hierarchy


def get_generation_order(
    class_linkage: pd.DataFrame, parent_col: str = "source_class", child_col: str = "target_class"
) -> dict:
    parent_child_list = class_linkage[[parent_col, child_col]].values.tolist()
    # Build a directed graph and a list of all names that have no parent
    graph: dict[str, set[str]] = {name: set() for tup in parent_child_list for name in tup}
    has_parent = {name: False for tup in parent_child_list for name in tup}
    for parent, child in parent_child_list:
        graph[parent].add(child)
        has_parent[child] = True

    # All names that have absolutely no parent:
    roots = [name for name, parents in has_parent.items() if not parents]

    return _traverse({}, graph, roots)


def prettify_generation_order(generation_order: dict, depth: dict | None = None, start=-1) -> dict:
    """Prettifies generation order dictionary for easier consumption."""
    depth = depth or {}
    for key, value in generation_order.items():
        depth[key] = start + 1
        if isinstance(value, dict):
            prettify_generation_order(value, depth, start=start + 1)
    return OrderedDict(sorted(depth.items(), key=lambda item: item[1]))


def epoch_now_ms():
    return int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds() * 1000)


def chunker(sequence, chunk_size):
    return [sequence[i : i + chunk_size] for i in range(0, len(sequence), chunk_size)]


def datetime_utc_now():
    return datetime.now(timezone.utc)


def retry_decorator(max_retries=2, retry_delay=3, component_name=""):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            previous_exception = None
            for attempt in range(max_retries + 1):
                try:
                    logging.debug(f"Attempt {attempt + 1} of {max_retries + 1} for {component_name}")
                    return func(*args, **kwargs)
                except CogniteReadTimeout as e:
                    previous_exception = e
                    if attempt < max_retries:
                        logging.error(
                            f"""CogniteReadTimeout retry attempt {attempt + 1} failed for {component_name} .
                            Retrying in {retry_delay} second(s). Error:"""
                        )
                        logging.error(e)
                        time.sleep(retry_delay)
                    else:
                        raise e
                except CogniteDuplicatedError as e:
                    if isinstance(previous_exception, CogniteReadTimeout):
                        # if previous exception was CogniteReadTimeout,
                        # we can't be sure if the items were created or not
                        if len(e.successful) == 0 and len(e.failed) == 0 and len(e.duplicated) >= 0:
                            logging.warning(
                                f"Duplicate error for {component_name} . All items already exist in CDF. "
                                "Suppressing error."
                            )
                            return
                        else:
                            # can happend because of eventual consistency. Retry with delay to allow for CDF to catch up
                            if attempt < max_retries:
                                logging.error(
                                    f"""CogniteDuplicatedError retry attempt {attempt + 1} failed for {component_name} .
                                      Retrying in {retry_delay} second(s). Error:"""
                                )
                                logging.error(e)
                                # incerasing delay to allow for CDF to catch up
                                time.sleep(retry_delay)
                            else:
                                raise e
                    else:
                        # no point in retrying duplicate error if previous exception was not a timeout
                        raise e

                except Exception as e:
                    previous_exception = e
                    if attempt < max_retries:
                        logging.error(
                            f"Retry attempt {attempt + 1} failed for {component_name}. "
                            f"Retrying in {retry_delay} second(s)."
                        )
                        logging.error(e)
                        time.sleep(retry_delay)
                    else:
                        raise e

        return wrapper

    return decorator


def create_sha256_hash(string: str) -> str:
    # Create a SHA-256 hash object
    sha256_hash = hashlib.sha256()

    # Convert the string to bytes and update the hash object
    sha256_hash.update(string.encode("utf-8"))

    # Get the hexadecimal representation of the hash
    hash_value = sha256_hash.hexdigest()

    return hash_value


def generate_exception_report(exceptions: list[dict], category: str = "") -> str:
    exceptions_as_dict = _order_expectations_by_type(exceptions) if exceptions else {}
    report = ""

    for exception_type in exceptions_as_dict.keys():
        title = f"# {category}: {exception_type}" if category else ""
        warnings = "\n- " + "\n- ".join(exceptions_as_dict[exception_type])
        report += title + warnings + "\n\n"

    return report


def _order_expectations_by_type(exceptions: list[dict]) -> dict[str, list[str]]:
    exception_dict: dict[str, list[str]] = {}
    for exception in exceptions:
        if exception["loc"]:
            location = f"[{'/'.join(exception['loc'])}]"
        else:
            location = ""

        issue = f"{exception['msg']} {location}"

        if exception_dict.get(exception["type"]) is None:
            exception_dict[exception["type"]] = [issue]
        else:
            exception_dict[exception["type"]].append(issue)
    return exception_dict



class NeatException(Exception):
    """Base class for all exceptions raised by NEAT."""

    ...


class NeatImportError(NeatException):
    """Cognite Import Error

    Raised if the user attempts to use functionality which requires an uninstalled package.

    Args:
        module (str): Name of the module which could not be imported
        extra (str): The name of the extra you use to install it with neat.
    """

    def __init__(self, module: str, extra: str):
        self.module = module
        self.message = (
            f"This functionality requires {self.module}. "
            f'You can include it in your neat installation with `pip install "cognite-neat[{extra}]"`.'
        )

    def __str__(self) -> str:
        return self.message


class InvalidWorkFlowError(NeatException):
    """InvalidWorkFlowError
    Raised if an invalid workflow is provided to the Workflow Manager.
    Args:
        step_name (str): Name of the step which could not be run
        missing_data (list[str]): The missing data for the step.
    """

    def __init__(self, step_name, missing_data: list[str]):
        self.message = f"In the workflow step {step_name} the following data is missing: {missing_data}."

    def __str__(self) -> str:
        return self.message



"""This module provides set of methods that perform conversion of TransformationRules
to TransformationRules for purpose of for example:

- subsetting the data model to only include desired classes and their properties
- converting classes/properties ids/names to DMS compliant format
"""
import logging
import re
import warnings

from cognite.neat.rules.analysis import get_defined_classes
from cognite.neat.rules.models import TransformationRules


def subset_rules(
    transformation_rules: TransformationRules, desired_classes: set, skip_validation: bool = False
) -> TransformationRules:
    """Subset transformation rules to only include desired classes and their properties.

    Parameters
    ----------
    transformation_rules : TransformationRules
        Instance of TransformationRules to subset
    desired_classes : set
        Desired classes to include in the reduced data model
    skip_validation : bool
        To skip underlying pydantic validation, by default False

    Returns
    -------
    TransformationRules
        Instance of TransformationRules

    Notes
    -----
    It is fine to skip validation since we are deriving the reduced data model from data
    model (i.e. TransformationRules) which has already been validated.
    """

    defined_classes = get_defined_classes(transformation_rules)
    possible_classes = defined_classes.intersection(desired_classes)
    impossible_classes = desired_classes - possible_classes

    if not possible_classes:
        logging.error("None of the desired classes are defined in the data model!")
        raise ValueError("None of the desired classes are defined in the data model!")

    if impossible_classes:
        logging.warning(f"Could not find the following classes defined in the data model: {impossible_classes}")
        warnings.warn(
            f"Could not find the following classes defined in the data model: {impossible_classes}", stacklevel=2
        )

    reduced_data_model = {
        "metadata": transformation_rules.metadata,
        "prefixes": transformation_rules.prefixes,
        "classes": {},
        "properties": {},
        "instances": transformation_rules.instances,
    }

    logging.info(f"Reducing data model to only include the following classes: {possible_classes}")
    for class_ in possible_classes:
        reduced_data_model["classes"][class_] = transformation_rules.classes[class_]

    for id_, property_definition in transformation_rules.properties.items():
        if property_definition.class_id in possible_classes:
            reduced_data_model["properties"][id_] = property_definition

    if skip_validation:
        return TransformationRules.construct(**reduced_data_model)
    else:
        return TransformationRules(**reduced_data_model)


def to_dms_compliant_rules(rules: TransformationRules) -> TransformationRules:
    ...


# to be used for conversion to DMS compliant format
def to_dms_name(name: str, entity_type: str, fix_casing: bool = False) -> str:
    """
    Repairs an entity name to conform to GraphQL naming convention
    >>> repair_name("wind-speed", "property")
    'windspeed'
    >>> repair_name("Wind.Speed", "property", True)
    'windSpeed'
    >>> repair_name("windSpeed", "class", True)
    'WindSpeed'
    >>> repair_name("22windSpeed", "class")
    '_22windSpeed'
    """

    # Remove any non GraphQL compliant characters
    repaired_string = re.sub(r"[^_a-zA-Z0-9]", "", name)

    # Name must start with a letter or underscore
    if repaired_string[0].isdigit():
        repaired_string = f"_{repaired_string}"

    if not fix_casing:
        return repaired_string
    # Property names must be camelCase
    if entity_type == "property" and repaired_string[0].isupper():
        return repaired_string[0].lower() + repaired_string[1:]
    # Class names must be PascalCase
    elif entity_type == "class" and repaired_string[0].islower():
        return repaired_string[0].upper() + repaired_string[1:]
    else:
        return repaired_string



import json
import logging
import os
import shutil
import time
import zipfile
from pathlib import Path
from typing import Dict

from cognite.client import CogniteClient
from cognite.client.data_classes import Event, FileMetadataUpdate, LabelDefinition, LabelFilter
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel

from cognite.neat.workflows.model import WorkflowFullStateReport, WorkflowState, WorkflowStepEvent
from cognite.neat.workflows.utils import get_file_hash


class NeatCdfResource(BaseModel):
    id: int = None
    name: str
    rtype: str
    last_updated_time: int = None
    last_updated_by: str = None
    version: str = None
    tag: str = None
    comments: str = None
    external_id: str = None
    is_latest: bool = False


class CdfStore:
    def __init__(
        self,
        client: CogniteClient,
        data_set_id: int,
        workflows_storage_path: Path = None,
        rules_storage_path: Path = None,
    ):
        self.client = client
        self.data_set_id = data_set_id
        # todo use pathlib
        self.workflows_storage_path = str(workflows_storage_path)
        self.rules_storage_path = str(rules_storage_path)
        self.workflows_storage_type = "file"

    def init_cdf_resources(self, resource_type="all"):
        if self.client and self.data_set_id:
            try:
                if resource_type == "all" or resource_type == "labels":
                    list = self.client.labels.list(external_id_prefix="neat-")
                    if len(list) == 0:
                        labels = [
                            LabelDefinition(
                                external_id="neat-workflow", name="neat-workflow", data_set_id=self.data_set_id
                            ),
                            LabelDefinition(
                                external_id="neat-latest", name="neat-latest", data_set_id=self.data_set_id
                            ),
                        ]
                        self.client.labels.create(labels)
                    else:
                        logging.debug("Labels already exists.")
            except Exception as e:
                logging.debug(f"Failed to create labels.{e}")

    def package_workflow(self, workflow_name):
        """Creates a zip archive from a folder"""
        folder_path = os.path.join(self.workflows_storage_path, workflow_name)
        archive_path = os.path.join(self.workflows_storage_path, f"{workflow_name}.zip")
        # Make sure the folder exists
        if not os.path.isdir(folder_path):
            logging.error(f"Error: {folder_path} is not a directory")
            raise Exception(f"Can't package workflow.Error: {folder_path} is not a directory")

        # Create a ZipFile object with write mode
        with zipfile.ZipFile(archive_path, "w", zipfile.ZIP_DEFLATED) as zipf:
            # Walk through the directory and add each json file to the archive
            for root, _, files in os.walk(folder_path):
                for file in files:
                    file_path = os.path.join(root, file)
                    # Check if the file has a .json extension
                    if (
                        file.endswith(".yaml")
                        or file.endswith(".py")
                        or file.endswith(".xlsx")
                        or file.endswith(".csv")
                        or file.endswith(".json")
                        or file.endswith(".parquet")
                    ):
                        # Add the file to the archive
                        zipf.write(file_path, os.path.relpath(file_path, folder_path))

    def extract_workflow_package(self, workflow_name: str):
        # Make sure the archive exists
        workflow_name = workflow_name.replace(".zip", "")
        package_full_path = Path(self.workflows_storage_path).joinpath(f"{workflow_name}.zip")
        output_folder = Path(self.workflows_storage_path).joinpath(workflow_name)
        if not os.path.isfile(package_full_path):
            print(f"Error: {package_full_path} is not a file")
            raise Exception(f"Can't extract workflow package. Error: {package_full_path} is not a file")

        # Remove the output folder if it already exists
        if output_folder.exists():
            shutil.rmtree(output_folder)
        # Create the output folder if it does not exist
        os.makedirs(output_folder, exist_ok=True)

        # Extract the contents of the archive to the output folder
        with zipfile.ZipFile(package_full_path, "r") as zipf:
            zipf.extractall(output_folder)

    def load_workflows_from_cfg_by_filter(self, config_filter: list[str] = None) -> str:
        """Load workflow package from CDF and extract it to the storage path"""
        # filter syntax: name:workflow_name=version; tag:tag_name
        try:
            for filter_item in config_filter:
                filter_type = filter_item.split(":")[0]
                if filter_type == "name":
                    filter_workflow_name_with_version_l = filter_item.split(":")[1].split("=")
                    filter_workflow_name = filter_workflow_name_with_version_l[0]
                    if len(filter_workflow_name_with_version_l) > 1:
                        filter_workflow_version = filter_workflow_name_with_version_l[1]
                    else:
                        filter_workflow_version = None
                    logging.info(
                        f"Loading workflow filtered by name: {filter_workflow_name} version: {filter_workflow_version}"
                    )
                    self.load_workflows_from_cdf(filter_workflow_name, filter_workflow_version)
                elif filter_type == "tag":
                    tag = filter_item.split(":")[1]
                    logging.info(f"Loading workflow filtered by tag: {tag}")
                    self.load_workflows_from_cdf(metadata_filter={"tag": tag})

        except Exception as e:
            logging.error(f"Failed to load workflows by filter. Error: {e}")

    def load_workflows_from_cdf(
        self, workflow_name: str = None, version: str = None, metadata_filter: Dict[str, str] = None
    ) -> list[str]:
        """Load workflow package or multiple workfllows from CDF and extract it to the storage path"""
        # TODO: Add workflow and tag filters
        if self.data_set_id and self.client:
            if workflow_name:
                workflow_name = workflow_name if ".zip" in workflow_name else f"{workflow_name}.zip"
            labels = ["neat-workflow"]
            meta = metadata_filter or {}
            if version:
                meta["hash"] = version
            else:
                labels.append("neat-latest")
            label_filter = LabelFilter(contains_all=labels)
            if workflow_name:
                files_metada = self.client.files.list(name=workflow_name, labels=label_filter, metadata=meta)
            elif metadata_filter:
                files_metada = self.client.files.list(labels=label_filter, metadata=meta)
            else:
                raise Exception("Workflow name or metadata_filter is required")

            files_external_ids = []
            for file_meta in files_metada:
                self.client.files.download(self.workflows_storage_path, external_id=file_meta.external_id)
                logging.info(
                    f"Workflow {file_meta.name} , "
                    f"external_id = {file_meta.external_id} syccessfully downloaded from CDF"
                )
                self.extract_workflow_package(file_meta.name)
                files_external_ids.append(file_meta.external_id)
            return files_external_ids
        else:
            logging.error("No CDF client or data set id provided")

    def save_workflow_to_cdf(self, name: str, tag: str = "", changed_by: str = "", comments: str = "") -> None:
        """Saves entire workflow (all artifacts) to CDF."""
        self.package_workflow(name)
        if self.data_set_id and self.client:
            zip_file = Path(self.workflows_storage_path) / f"{name}.zip"
            if zip_file.exists():
                self.save_resource_to_cdf(name, "workflow-package", zip_file, tag, changed_by, comments)
                os.remove(zip_file)
                logging.info(f"Workflow package {name} uploaded to CDF")
            else:
                logging.error(f"Workflow package {name} not found, skipping")
        else:
            logging.error("No CDF client or data set id provided")

    def save_resource_to_cdf(
        self,
        workflow_name: str,
        resource_type: str,
        file_path: Path,
        tag: str = "",
        changed_by: str = "",
        comments: str = "",
    ) -> None:
        """Saves resources to cdf. For instance workflow package, rules file, etc"""
        if self.data_set_id and self.client:
            if not os.path.exists(file_path):
                logging.error(f"File {file_path} not found, skipping")
                return
            # removing the latest tag from all files related to this workflow
            files_metada = self.client.files.list(
                name=file_path.name,
                metadata={
                    "workflow_name": workflow_name,
                    "resource_type": resource_type,
                },
            )
            for file_meta in files_metada:
                meta_update = FileMetadataUpdate(id=file_meta.id).labels.remove("neat-latest")
                self.client.files.update(meta_update)

            hash_ = get_file_hash(file_path)
            self.client.files.upload(
                file_path,
                name=file_path.name,
                external_id=f"neat-wf-{hash_}",
                labels=["neat-workflow", "neat-latest"],
                metadata={
                    "tag": tag,
                    "hash": hash_,
                    "workflow_name": workflow_name,
                    "resource_type": resource_type,
                    "changed_by": changed_by,
                    "comments": comments,
                },
                source="neat",
                overwrite=True,
                data_set_id=self.data_set_id,
            )
        else:
            logging.error("No CDF client or data set id provided")

    def load_rules_file_from_cdf(self, name: str, version: str = None):
        logging.info(f"Loading rules file {name} (version = {version} ) from CDF ")
        # TODO: Download latest if version is not specified
        files_metadata = self.client.files.list(name=name, metadata={"hash": version})
        if len(files_metadata) > 0:
            self.client.files.download(self.rules_storage_path, external_id=files_metadata[0].external_id)
        else:
            raise Exception(f"Rules file {name} with version {version} not found in CDF")

    def get_list_of_resources_from_cdf(self, resource_type: str) -> list[NeatCdfResource]:
        logging.info(f"Getting list of resources of type {resource_type} from CDF")
        label_filter = LabelFilter(contains_any=["neat-workflow"])
        files_metadata = self.client.files.list(labels=label_filter, metadata={"resource_type": resource_type})

        return [
            NeatCdfResource(
                id=file_meta.id,
                name=file_meta.name,
                external_id=file_meta.external_id,
                version=file_meta.metadata["hash"],
                rtype=file_meta.metadata["resource_type"],
                comments=file_meta.metadata["comments"],
                last_updated_by=file_meta.metadata["changed_by"],
                tag=file_meta.metadata["tag"],
                last_updated_time=file_meta.last_updated_time,
            )
            for file_meta in files_metadata
        ]

    def get_list_of_workflow_executions_from_cdf(self, limit=200) -> list[WorkflowFullStateReport]:
        """Returns list of workflow executions from CDF."""
        logging.debug("Getting list of workflow executions from CDF")
        events = self.client.events.list(type="neat-workflow-run", source="neat", sort=["startTime:desc"], limit=limit)
        executions = []
        for event in events:
            if (time.time() - event.start_time > 24 * 60 * 60) and event.subtype == "RUNNING":
                event.subtype = "EXPIRED"

            try:
                executions.append(
                    WorkflowFullStateReport(
                        run_id=event.metadata["run_id"],
                        elapsed_time=float(event.metadata["elapsed_time"]) if "elapsed_time" in event.metadata else 0,
                        last_error=event.metadata["error"] if "error" in event.metadata else "",
                        execution_log=[],
                        workflow_name=event.metadata["workflow_name"] if "workflow_name" in event.metadata else "",
                        last_updated_time=event.last_updated_time,
                        start_time=event.start_time,
                        end_time=event.end_time,
                        state=event.subtype,
                    )
                )
            except Exception as e:
                logging.info(
                    f"Failed to parse workflow execution event for workflow, "
                    f"run_id = {event.metadata['run_id']}, error = {e}"
                )
        return executions

    def get_detailed_workflow_execution_report_from_cdf(self, run_id: str) -> WorkflowFullStateReport:
        """Returns detailed workflow execution report from CDF"""
        logging.debug(f"Getting detailed workflow execution {run_id} from CDF")
        external_id = f"neat-wf-run-{run_id}"
        event = self.client.events.retrieve(external_id=external_id)
        steps_log = []
        if event:
            try:
                if "execution_log" in event.metadata:
                    steps_log = json.loads(event.metadata["execution_log"])
                else:
                    external_id = f"neat-wf-execution-log-{run_id}"
                    steps_log = self.client.files.download_bytes(external_id=external_id)
            except Exception as e:
                logging.info(
                    f"Failed to parse execution log for workflow {event.metadata['workflow_name']}, "
                    f"run_id = {event.metadata['run_id']}, error = {e}"
                )
                steps_log = []
            return WorkflowFullStateReport(
                run_id=event.metadata["run_id"],
                elapsed_time=float(event.metadata["elapsed_time"]) if "elapsed_time" in event.metadata else 0,
                last_error=event.metadata["error"] if "error" in event.metadata else "",
                execution_log=steps_log,
                workflow_name=event.metadata["workflow_name"] if "workflow_name" in event.metadata else "",
                last_updated_time=event.last_updated_time,
                start_time=event.start_time,
                end_time=event.end_time,
                state=event.subtype,
            )
        raise Exception(f"Workflow execution with run_id = {run_id} not found")

    def report_workflow_execution_to_cdf(self, report: WorkflowFullStateReport):
        # Report workflow run to CDF as single event with all the steps attached either in metadata
        # or attached file (depends on the size)
        metadata = {
            "run_id": report.run_id,
            "elapsed_time": report.elapsed_time,
            "error": report.last_error,
            "execution_log": "",
            "workflow_name": report.workflow_name,
        }
        external_id = f"neat-wf-run-{report.run_id}"
        # serialized_execution_log = json.dumps([step.dict() for step in report.execution_log])
        serialized_execution_log = json.dumps(jsonable_encoder(report.execution_log))
        execution_log_size = bytes(serialized_execution_log, "utf-8").__sizeof__()
        logging.debug(f"Serialized execution log size: {execution_log_size}")
        # logging.debug(f"Serialized execution log: {serialized_execution_log}")

        if report.start_time:
            report.start_time = report.start_time * 1000
        if report.end_time:
            report.end_time = report.end_time * 1000

        event = Event(
            external_id=external_id,
            type="neat-workflow-run",
            subtype=report.state,
            start_time=report.start_time,
            end_time=report.end_time,
            source="neat",
            metadata=metadata,
            data_set_id=self.data_set_id,
        )

        if execution_log_size > 128000:  # 128K limit for Event metadata
            self.client.files.upload_bytes(
                serialized_execution_log,
                name="neat-workflow-execution-log.json",
                external_id=f"neat-wf-execution-log-{report.run_id}",
                metadata={"run_id": report.run_id, "workflow_name": report.workflow_name},
                source="neat",
                overwrite=True,
                data_set_id=self.data_set_id,
            )
        elif serialized_execution_log:
            event.metadata["execution_log"] = serialized_execution_log

        if report.state in [WorkflowState.CREATED, WorkflowState.RUNNING]:
            res = self.client.events.create(event)
        elif report.state in [WorkflowState.COMPLETED, WorkflowState.FAILED]:
            res = self.client.events.update(event)
        else:
            logging.error(f"Unknown workflow state {report.state}")
            return

        if res.external_id == external_id:
            logging.debug(f"Workflow run {report.run_id} reported to CDF")
        else:
            logging.error(f"Workflow run {report.run_id} not reported to CDF due to error: {res}")

    def report_step_event_to_cdf(self, event: WorkflowStepEvent):
        pass



import shutil
from typing import Any, Dict, Generator, Iterable, Iterator, Mapping, Optional, Tuple, Union

import pyoxigraph as ox
from rdflib import Graph
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.plugins.sparql.sparql import Query, Update
from rdflib.query import Result
from rdflib.store import VALID_STORE, Store
from rdflib.term import BNode, Identifier, Literal, Node, URIRef, Variable

__all__ = ["OxigraphStore"]

_Triple = Tuple[Node, Node, Node]
_Quad = Tuple[Node, Node, Node, Graph]
_TriplePattern = Tuple[Optional[Node], Optional[Node], Optional[Node]]


class OxigraphStore(Store):
    context_aware: bool = True
    formula_aware: bool = False
    transaction_aware: bool = False
    graph_aware: bool = True

    def __init__(
        self,
        configuration: Optional[str] = None,
        identifier: Optional[Identifier] = None,
        *,
        store: Optional[ox.Store] = None,
    ):
        self._store = store
        self._prefix_for_namespace: Dict[URIRef, str] = {}
        self._namespace_for_prefix: Dict[str, URIRef] = {}
        super().__init__(configuration, identifier)

    def open(self, configuration: str, create: bool = False) -> Optional[int]:  # noqa: ARG002
        if self._store is not None:
            raise ValueError("The open function should be called before any RDF operation")
        self._store = ox.Store(configuration)
        return VALID_STORE

    def close(self, commit_pending_transaction: bool = False) -> None:  # noqa: ARG002
        del self._store

    def destroy(self, configuration: str) -> None:
        shutil.rmtree(configuration)

    def gc(self) -> None:
        pass

    @property
    def _inner(self) -> ox.Store:
        if self._store is None:
            self._store = ox.Store()
        return self._store

    def add(
        self,
        triple: _Triple,
        context: Graph,
        quoted: bool = False,
    ) -> None:
        if quoted:
            raise ValueError("Oxigraph stores are not formula aware")
        self._inner.add(_to_ox(triple, context))
        super().add(triple, context, quoted)

    def addN(self, quads: Iterable[_Quad]) -> None:  # noqa: N802
        self._inner.extend([_to_ox(q) for q in quads])
        for quad in quads:
            (s, p, o, g) = quad
            super().add((s, p, o), g)

    def remove(
        self,
        triple: _TriplePattern,
        context: Optional[Graph] = None,
    ) -> None:
        for q in self._inner.quads_for_pattern(*_to_ox_quad_pattern(triple, context)):
            self._inner.remove(q)
        super().remove(triple, context)

    def triples(
        self,
        triple_pattern: _TriplePattern,
        context: Optional[Graph] = None,
    ) -> Iterator[Tuple[_Triple, Iterator[Optional[Graph]]]]:
        return (_from_ox(q) for q in self._inner.quads_for_pattern(*_to_ox_quad_pattern(triple_pattern, context)))

    def __len__(self, context: Optional[Graph] = None) -> int:
        if context is None:
            # TODO: very bad
            return len({q.triple for q in self._inner})
        return sum(1 for _ in self._inner.quads_for_pattern(None, None, None, _to_ox(context)))

    def contexts(self, triple: Optional[_Triple] = None) -> Generator[Graph, None, None]:
        if triple is None:
            return (_from_ox(g) for g in self._inner.named_graphs())
        return (_from_ox(q[3]) for q in self._inner.quads_for_pattern(*_to_ox_quad_pattern(triple)))

    def query(
        self,
        query: Union[Query, str],
        initNs: Mapping[str, Any],  # noqa: N803
        initBindings: Mapping[str, Identifier],  # noqa: N803
        queryGraph: str,  # noqa: N803
        **kwargs: Any,
    ) -> "Result":
        if isinstance(queryGraph, Query) or kwargs:
            raise NotImplementedError
        # queryGraph = str(queryGraph)
        init_ns = dict(self._namespace_for_prefix, **initNs)
        query = "".join(f"PREFIX {prefix}: <{namespace}>\n" for prefix, namespace in init_ns.items()) + query
        if initBindings:
            query += "\nVALUES ( {} ) {{ ({}) }}".format(
                " ".join(f"?{k}" for k in initBindings), " ".join(v.n3() for v in initBindings.values())
            )
        result = self._inner.query(
            query,
            use_default_graph_as_union=queryGraph == "__UNION__",
            default_graph=_to_ox(queryGraph) if isinstance(queryGraph, Node) else None,
        )
        if isinstance(result, bool):
            out = Result("ASK")
            out.askAnswer = result
        elif isinstance(result, ox.QuerySolutions):
            out = Result("SELECT")
            out.vars = [Variable(v.value) for v in result.variables]
            out.bindings = ({v: _from_ox(val) for v, val in zip(out.vars, solution)} for solution in result)
        elif isinstance(result, ox.QueryTriples):
            out = Result("CONSTRUCT")
            out.graph = Graph()
            out.graph += (_from_ox(t) for t in result)
        else:
            raise ValueError(f"Unexpected query result: {result}")
        return out

    def update(
        self,
        update: Union[Update, str],
        initNs: Mapping[str, Any],  # noqa: N803
        initBindings: Mapping[str, Identifier],  # noqa: N803
        queryGraph: str,  # noqa: N803
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    def commit(self) -> None:
        # TODO: implement
        pass

    def rollback(self) -> None:
        # TODO: implement
        pass

    def add_graph(self, graph: Graph) -> None:
        self._inner.add_graph(_to_ox(graph))

    def remove_graph(self, graph: Graph) -> None:
        self._inner.remove_graph(_to_ox(graph))

    def bind(self, prefix: str, namespace: URIRef, override: bool = True) -> None:
        if not override and (prefix in self._namespace_for_prefix or namespace in self._prefix_for_namespace):
            return  # nothing to do
        self._delete_from_prefix(prefix)
        self._delete_from_namespace(namespace)
        self._namespace_for_prefix[prefix] = namespace
        self._prefix_for_namespace[namespace] = prefix

    def _delete_from_prefix(self, prefix):
        if prefix not in self._namespace_for_prefix:
            return
        namespace = self._namespace_for_prefix[prefix]
        del self._namespace_for_prefix[prefix]
        self._delete_from_namespace(namespace)

    def _delete_from_namespace(self, namespace):
        if namespace not in self._prefix_for_namespace:
            return
        prefix = self._prefix_for_namespace[namespace]
        del self._prefix_for_namespace[namespace]
        self._delete_from_prefix(prefix)

    def prefix(self, namespace: URIRef) -> Optional[str]:
        return self._prefix_for_namespace.get(namespace)

    def namespace(self, prefix: str) -> Optional[URIRef]:
        return self._namespace_for_prefix.get(prefix)

    def namespaces(self) -> Iterator[Tuple[str, URIRef]]:
        yield from self._namespace_for_prefix.items()


def _to_ox(term: Union[Node, _Triple, _Quad, Graph], context: Optional[Graph] = None):
    if term is None:
        return None
    if term == DATASET_DEFAULT_GRAPH_ID:
        return ox.DefaultGraph()
    if isinstance(term, URIRef):
        return ox.NamedNode(term)
    if isinstance(term, BNode):
        return ox.BlankNode(term)
    if isinstance(term, Literal):
        return ox.Literal(term, language=term.language, datatype=ox.NamedNode(term.datatype) if term.datatype else None)
    if isinstance(term, Graph):
        return _to_ox(term.identifier)
    if isinstance(term, tuple):
        if len(term) == 3:
            return ox.Quad(_to_ox(term[0]), _to_ox(term[1]), _to_ox(term[2]), _to_ox(context))
        if len(term) == 4:
            return ox.Quad(_to_ox(term[0]), _to_ox(term[1]), _to_ox(term[2]), _to_ox(term[3]))
    raise ValueError(f"Unexpected rdflib term: {repr(term)}")


def _to_ox_quad_pattern(triple: _TriplePattern, context: Optional[Graph] = None):
    (s, p, o) = triple
    return _to_ox_term_pattern(s), _to_ox_term_pattern(p), _to_ox_term_pattern(o), _to_ox_term_pattern(context)


def _to_ox_term_pattern(term):
    if term is None:
        return None
    if isinstance(term, URIRef):
        return ox.NamedNode(term)
    if isinstance(term, BNode):
        return ox.BlankNode(term)
    if isinstance(term, Literal):
        return ox.Literal(term, language=term.language, datatype=ox.NamedNode(term.datatype) if term.datatype else None)
    if isinstance(term, Graph):
        return _to_ox(term.identifier)
    raise ValueError(f"Unexpected rdflib term: {repr(term)}")


def _from_ox(term):
    if term is None:
        return None
    if isinstance(term, ox.NamedNode):
        return URIRef(term.value)
    if isinstance(term, ox.BlankNode):
        return BNode(term.value)
    if isinstance(term, ox.Literal):
        if term.language:
            return Literal(term.value, lang=term.language)
        return Literal(term.value, datatype=URIRef(term.datatype.value))
    if isinstance(term, ox.DefaultGraph):
        return None
    if isinstance(term, ox.Triple):
        return _from_ox(term.subject), _from_ox(term.predicate), _from_ox(term.object)
    if isinstance(term, ox.Quad):
        return (_from_ox(term.subject), _from_ox(term.predicate), _from_ox(term.object)), _from_ox(term.graph_name)
    raise ValueError(f"Unexpected Oxigraph term: {repr(term)}")



from cognite.neat.workflows.base import BaseWorkflow
from .model import WorkflowStepDefinition, FlowMessage, WorkflowStepEvent, WorkflowFullStateReport
from .manager import WorkflowManager

__all__ = [
    "BaseWorkflow",
    "WorkflowStepDefinition",
    "WorkflowManager",
    "WorkflowStepEvent",
    "WorkflowFullStateReport",
    "FlowMessage",
]



import importlib
import inspect
import os
from pathlib import Path
import sys
import types

from pydantic import BaseModel
from cognite.neat.app.monitoring.metrics import NeatMetricsCollector
from cognite.neat.exceptions import InvalidWorkFlowError
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigItem, WorkflowConfigs
import cognite.neat.workflows.steps.lib
import logging

from cognite.neat.workflows.steps.step_model import DataContract, T_Output
from cognite.neat.workflows.steps.step_model import Step


class StepMetadata(BaseModel):
    name: str
    description: str = ""
    category: str = "default"  # defines the category the step belongs to (e.g. "data", "model", "test")
    scope: str = "global"  # defines the scope of the step (e.g. "global", local to a specific workflow)
    input: list[str]
    output: list[str]
    configuration_templates: list[WorkflowConfigItem] = []


class StepsRegistry:
    def __init__(self, data_store_path: str | None = None):
        self._step_classes: list[Step] = []
        self.user_steps_path = Path(data_store_path) / "steps"
        self.data_store_path = data_store_path

    def load_step_classes(self):
        if self._step_classes:
            # classes already loaded - no need to reload
            return
        for name, step_cls in inspect.getmembers(cognite.neat.workflows.steps.lib):
            if inspect.isclass(step_cls):
                logging.info(f"Loading NEAT step {name}")
                self._step_classes.append(step_cls)
        try:
            if self.user_steps_path:
                self.load_custom_step_classes(self.user_steps_path, scope="global")
        except Exception:
            logging.info(f"No user defined modules provided in {self.user_steps_path}")

    def load_custom_step_classes(self, custom_steps_path: Path, scope: str = "global"):
        sys.path.append(str(custom_steps_path))
        for step_module_name in os.listdir(custom_steps_path):
            step_module_path = custom_steps_path / Path(step_module_name)
            if step_module_name.startswith("__") or (
                step_module_path.is_file() and not step_module_name.endswith(".py")
            ):
                continue

            steps_module = importlib.import_module(step_module_name.replace(".py", ""))
            logging.info(f"Loading user defined step from {steps_module}")
            for name, step_cls in inspect.getmembers(steps_module):
                base_class = getattr(step_cls, "__bases__", None)
                if (
                    name.startswith("__")
                    or base_class is None
                    or len(base_class) == 0
                    or base_class[0].__name__ != "Step"
                ):
                    continue
                logging.info(f"Loading user defined step {name} from {steps_module}")
                if inspect.isclass(step_cls):
                    logging.info(f"Loading user defined step {name} class {step_cls.__name__}")
                    step_cls.scope = scope
                    self._step_classes.append(step_cls)

    def run_step(
        self,
        step_name: str,
        flow_context: dict[str, DataContract],
        metrics: NeatMetricsCollector = None,
        configs: WorkflowConfigs = None,
    ) -> T_Output | None:
        for step_cls in self._step_classes:
            if step_cls.__name__ == step_name:
                step_obj: Step = step_cls(self.data_store_path)
                step_obj.set_flow_context(flow_context)
                step_obj.set_metrics(metrics)
                step_obj.set_workflow_configs(configs)
                signature = inspect.signature(step_obj.run)
                parameters = signature.parameters
                is_valid = True
                input_data = []
                missing_data = []
                for parameter in parameters.values():
                    try:
                        if parameter.annotation is FlowMessage:
                            input_data.append(flow_context["FlowMessage"])
                        else:
                            input_data.append(flow_context[parameter.annotation.__name__])
                    except KeyError:
                        is_valid = False
                        logging.error(f"Missing data for step {step_name} parameter {parameter.name}")
                        missing_data.append(parameter.annotation.__name__)
                        continue
                if not is_valid:
                    raise InvalidWorkFlowError(step_name, missing_data)
                return step_obj.run(*input_data)

    def get_list_of_steps(self):
        steps: list[StepMetadata] = []
        for step_cls in self._step_classes:
            try:
                signature = inspect.signature(step_cls.run)
                parameters = signature.parameters
                input_data = []
                output_data = []
                for parameter in parameters.values():
                    if parameter.annotation.__name__ != "_empty":
                        input_data.append(parameter.annotation.__name__)
                return_annotation = signature.return_annotation
                if return_annotation:
                    if isinstance(return_annotation, types.UnionType):
                        for annotation in return_annotation.__args__:
                            output_data.append(annotation.__name__)
                    elif isinstance(return_annotation, tuple):
                        for annotation in return_annotation:
                            output_data.append(annotation.__name__)
                    else:
                        output_data.append(return_annotation.__name__)
                # logging.info(f"Step {name} output data: {return_annotation}")
                steps.append(
                    StepMetadata(
                        name=step_cls.__name__,
                        scope=step_cls.scope,
                        input=input_data,
                        description=step_cls.description,
                        category=step_cls.category,
                        output=output_data,
                        configuration_templates=step_cls.configuration_templates,
                    )
                )
            except AttributeError as e:
                logging.error(f"Step {step_cls.__name__} does not have a run method.Error: {e}")

        return steps



import contextlib
import logging
import time
from pathlib import Path

from prometheus_client import Gauge

from cognite.client import CogniteClient
from cognite.client.data_classes import AssetFilter
from cognite.neat.constants import PREFIXES

from cognite.neat.rules.models import TransformationRules
from cognite.neat.rules.parser import parse_rules_from_excel_file
from cognite.neat.rules.exporter.rules2triples import get_instances_as_triples

from cognite.neat.graph.extractors import NeatGraphStore, drop_graph_store
from cognite.neat.graph.transformations.transformer import RuleProcessingReport, domain2app_knowledge_graph
from cognite.neat.graph.loaders.core.labels import upload_labels
from cognite.neat.graph.loaders.core.rdf_to_assets import categorize_assets, rdf2assets, upload_assets
from cognite.neat.graph.loaders.core.rdf_to_relationships import (
    categorize_relationships,
    rdf2relationships,
    upload_relationships,
)
from cognite.neat.graph.loaders.validator import validate_asset_hierarchy

from cognite.neat.workflows import utils
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.base import BaseWorkflow
from cognite.neat.workflows.cdf_store import CdfStore


with contextlib.suppress(ValueError):
    prom_cdf_resource_stats = Gauge(
        "neat_graph_to_asset_hierarchy_wf_cdf_resource_stats",
        "CDF resource stats before and after running fast_graph workflow",
        ["resource_type", "state"],
    )
with contextlib.suppress(ValueError):
    prom_data_issues_stats = Gauge("neat_graph_to_asset_hierarchy_wf_data_issues", "Data validation issues", ["type"])


class GraphsAndRulesBaseWorkflow(BaseWorkflow):
    def __init__(self, name: str, client: CogniteClient):
        super().__init__(name, client, [])
        self.dataset_id: int = 0
        self.source_graph: NeatGraphStore = None
        self.solution_graph: NeatGraphStore = None
        self.transformation_rules: TransformationRules = None
        self.graph_source_type = "memory"

    def step_load_transformation_rules(self, flow_msg: FlowMessage = None):
        # Load rules from file or remote location
        cdf_store = CdfStore(self.cdf_client, self.dataset_id, rules_storage_path=self.rules_storage_path)

        rules_file = self.get_config_item("rules.file").value
        rules_file_path = Path(self.rules_storage_path, rules_file)
        version = self.get_config_item("rules.version").value

        if rules_file_path.exists() and not version:
            logging.info(f"Loading rules from {rules_file_path}")
        elif rules_file_path.exists() and version:
            hash = utils.get_file_hash(rules_file_path)
            if hash != version:
                cdf_store.load_rules_file_from_cdf(rules_file, version)
        else:
            cdf_store.load_rules_file_from_cdf(self.cdf_client, version)

        self.transformation_rules = parse_rules_from_excel_file(rules_file_path)
        self.dataset_id = self.transformation_rules.metadata.data_set_id
        logging.info(f"Loaded prefixes {str(self.transformation_rules.prefixes)} rules from {rules_file_path.name!r}.")
        output_text = f"Loaded {len(self.transformation_rules.properties)} rules"
        logging.info(output_text)
        return FlowMessage(output_text=output_text)

    def step_configuring_stores(self, flow_msg: FlowMessage = None, clean_start: bool = True):
        # Initialize source and solution graph stores . clean_start=True will delete all
        # artifacts(files , locks , etc) from previous runs
        logging.info("Initializing source graph")
        self.graph_source_type = self.get_config_item_value("source_rdf_store.type", self.graph_source_type)
        source_store_dir = self.get_config_item_value("source_rdf_store.disk_store_dir", None)
        solution_store_dir = self.get_config_item_value("solution_rdf_store.disk_store_dir", None)
        source_store_dir = Path(self.data_store_path) / Path(source_store_dir) if source_store_dir else None
        solution_store_dir = Path(self.data_store_path) / Path(solution_store_dir) if solution_store_dir else None
        logging.info(f"source_store_dir={source_store_dir}")
        logging.info(f"solution_store_dir={solution_store_dir}")
        if clean_start:
            drop_graph_store(self.source_graph, source_store_dir, force=True)
            drop_graph_store(self.solution_graph, solution_store_dir, force=True)

        self.source_graph = NeatGraphStore(
            prefixes=self.transformation_rules.prefixes, base_prefix="neat", namespace=PREFIXES["neat"]
        )

        if self.get_config_item_value("source_rdf_store.type"):
            self.source_graph.init_graph(
                self.get_config_item_value("source_rdf_store.type", self.graph_source_type),
                self.get_config_item_value("source_rdf_store.query_url"),
                self.get_config_item_value("source_rdf_store.update_url"),
                "neat-tnt",
                internal_storage_dir=source_store_dir,
            )

        if self.get_config_item_value("solution_rdf_store.type"):
            self.solution_graph = NeatGraphStore(
                prefixes=self.transformation_rules.prefixes, base_prefix="neat", namespace=PREFIXES["neat"]
            )

            self.solution_graph.init_graph(
                self.get_config_item_value("solution_rdf_store.type"),
                self.get_config_item_value("solution_rdf_store.query_url"),
                self.get_config_item_value("solution_rdf_store.update_url"),
                "tnt-solution",
                internal_storage_dir=solution_store_dir,
            )

        self.solution_graph.graph_db_rest_url = self.get_config_item_value("solution_rdf_store.api_root_url")
        return

    def step_load_source_graph(self, flow_msg: FlowMessage = None):
        # Load graph into memory or GraphDB
        if self.graph_source_type.lower() == "graphdb":
            try:
                result = self.source_graph.query("SELECT DISTINCT ?class WHERE { ?s a ?class }")
            except Exception as e:
                logging.error(f"Failed to query most likely remote DB is not running {e}")
                raise Exception("Failed to query graph , most likely remote DB is not running") from e
            else:
                logging.info(f"Loaded {len(result.bindings)} classes")
        elif self.graph_source_type.lower() in ("memory", "oxigraph"):
            if source_file := self.get_config_item_value("source_rdf_store.file"):
                graphs = Path(self.data_store_path) / "source-graphs"
                self.source_graph.import_from_file(graphs / source_file)
                logging.info(f"Loaded {source_file} into source graph.")
            else:
                raise ValueError("You need a source_rdf_store.file specified for source_rdf_store.type=memory")
        else:
            raise NotImplementedError(f"Graph type {self.graph_source_type} is not supported.")

        self.solution_graph.drop()
        return

    def step_run_transformation(self, flow_msg: FlowMessage = None):
        report = RuleProcessingReport()
        # run transformation and generate new graph
        self.solution_graph.set_graph(
            domain2app_knowledge_graph(
                self.source_graph.get_graph(),
                self.transformation_rules,
                app_instance_graph=self.solution_graph.get_graph(),
                extra_triples=get_instances_as_triples(self.transformation_rules),
                client=self.cdf_client,
                cdf_lookup_database=None,  # change this accordingly!
                processing_report=report,
            )
        )
        return FlowMessage(
            output_text=f"Total processed rules: { report.total_rules } , success: { report.total_success } , \
             no results: { report.total_success_no_results } , failed: { report.total_failed }",
            payload=report,
        )


class Graph2AssetHierarchyBaseWorkflow(GraphsAndRulesBaseWorkflow):
    def __init__(self, name: str, client: CogniteClient):
        super().__init__(name, client)
        self.dataset_id: int = 0
        self.stop_on_error = False
        self.count_create_assets = 0

    def step_create_cdf_labels(self, flow_msg: FlowMessage = None):
        logging.info("Creating CDF labels")
        upload_labels(self.cdf_client, self.transformation_rules, extra_labels=["non-historic", "historic"])

    def step_prepare_cdf_assets(self, flow_msg: FlowMessage):
        # export graph into CDF
        # TODO : decide on error handling and retry logic\

        rdf_asset_dicts = rdf2assets(
            self.solution_graph,
            self.transformation_rules,
            stop_on_exception=self.stop_on_error,
        )

        if not self.cdf_client:
            logging.info("Dry run, no CDF client available")
            return

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_before = self.cdf_client.assets.aggregate(
            filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
        )[0]["count"]

        prom_cdf_resource_stats.labels(resource_type="asset", state="count_before_neat_update").set(total_assets_before)
        logging.info(f"Total count of assets in CDF before upload: { total_assets_before }")

        orphan_assets, circular_assets = validate_asset_hierarchy(rdf_asset_dicts)

        prom_data_issues_stats.labels(type="circular_assets").set(len(circular_assets))
        prom_data_issues_stats.labels(type="orphan_assets").set(len(orphan_assets))

        if orphan_assets:
            logging.error(f"Found orphaned assets: {', '.join(orphan_assets)}")

            orphanage_asset_external_id = (
                f"{self.transformation_rules.metadata.externalIdPrefix}orphanage-{self.transformation_rules.metadata.data_set_id}"
                if self.transformation_rules.metadata.externalIdPrefix
                else "orphanage"
            )

            # Kill the process if you dont have orphanage asset in your asset hierarchy
            # and inform the user that it is missing !
            if orphanage_asset_external_id not in rdf_asset_dicts:
                msg = f"You dont have Orphanage asset {orphanage_asset_external_id} in asset hierarchy!"
                logging.error(msg)
                raise Exception(msg)

            logging.error("Orphaned assets will be assigned to 'Orphanage' root asset")

            for external_id in orphan_assets:
                rdf_asset_dicts[external_id]["parent_external_id"] = orphanage_asset_external_id

            orphan_assets, circular_assets = validate_asset_hierarchy(rdf_asset_dicts)

            logging.info(orphan_assets)
        else:
            logging.info("No orphaned assets found, your assets look healthy !")

        if circular_assets:
            msg = f"Found circular dependencies: {str(circular_assets)}"
            logging.error(msg)
            raise Exception(msg)
        elif orphan_assets:
            msg = f"Not able to fix orphans: {', '.join(orphan_assets)}"
            logging.error(msg)
            raise Exception(msg)
        else:
            logging.info("No circular dependency among assets found, your assets hierarchy look healthy !")

        self.categorized_assets, report = categorize_assets(
            self.cdf_client, rdf_asset_dicts, self.dataset_id, return_report=True
        )

        count_create_assets = len(self.categorized_assets["create"])
        count_update_assets = len(self.categorized_assets["update"])
        count_decommission_assets = len(self.categorized_assets["decommission"])
        count_resurrect_assets = len(self.categorized_assets["resurrect"])

        prom_cdf_resource_stats.labels(resource_type="asset", state="create").set(count_create_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="update").set(count_update_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="decommission").set(count_decommission_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="resurrect").set(count_resurrect_assets)

        self.count_create_assets = count_create_assets

        logging.info(f"Total count of assets to be created: { count_create_assets }")
        logging.info(f"Total count of assets to be updated: { count_update_assets }")
        logging.info(f"Total count of assets to be decommission: { count_decommission_assets }")
        logging.info(f"Total count of assets to be resurrect: { count_resurrect_assets }")

        msg = f"Total count of assets { len(rdf_asset_dicts) } of which: { count_create_assets } to be created"
        msg += f", { count_update_assets } to be updated"
        msg += f", { count_decommission_assets } to be decommissioned"
        msg += f", { count_resurrect_assets } to be resurrected"
        number_of_updates = len(report["decommission"])
        logging.info(f"Total number of updates: {number_of_updates}")
        return FlowMessage(output_text=msg)

    def step_upload_cdf_assets(self, flow_msg: FlowMessage = None):
        if flow_msg and flow_msg.payload and "action" in flow_msg.payload:
            if flow_msg.payload["action"] != "approve":
                raise Exception("Update not approved")

        if not self.cdf_client:
            logging.error("No CDF client available")
            raise Exception("No CDF client available")

        upload_assets(self.cdf_client, self.categorized_assets, max_retries=2, retry_delay=4)
        for _ in range(1000):
            total_assets_after = self.cdf_client.assets.aggregate(
                filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
            )[0]["count"]
            if total_assets_after >= self.count_create_assets:
                break
            logging.info(f"Waiting for assets to be created, current count {total_assets_after}")
            time.sleep(2)

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_after = self.cdf_client.assets.aggregate(
            filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
        )[0]["count"]

        prom_cdf_resource_stats.labels(resource_type="asset", state="count_after_neat_update").set(total_assets_after)
        logging.info(f"Total count of assets in CDF after update: { total_assets_after }")
        self.categorized_assets = None  # free up memory after upload .
        return FlowMessage(output_text=f"Total count of assets in CDF after update: { total_assets_after }")

    def step_prepare_cdf_relationships(self, flow_msg: FlowMessage = None):
        # create, categorize and upload relationships
        rdf_relationships = rdf2relationships(
            self.solution_graph.get_graph(),
            self.transformation_rules,
        )
        if not self.cdf_client:
            logging.info("Dry run, no CDF client available")
            return

        self.categorized_relationships = categorize_relationships(self.cdf_client, rdf_relationships, self.dataset_id)
        count_defined_relationships = len(rdf_relationships)
        count_create_relationships = len(self.categorized_relationships["create"])
        count_decommission_relationships = len(self.categorized_relationships["decommission"])
        count_resurrect_relationships = len(self.categorized_relationships["resurrect"])

        prom_cdf_resource_stats.labels(resource_type="relationships", state="defined").set(count_defined_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="create").set(count_create_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="decommission").set(
            count_decommission_relationships
        )
        prom_cdf_resource_stats.labels(resource_type="relationships", state="resurrect").set(
            count_resurrect_relationships
        )

        msg = (
            f"Total count of relationships { count_defined_relationships } of which:"
            f" { count_create_relationships } to be created"
        )
        msg += f", { count_decommission_relationships } to be decommissioned"
        msg += f", { count_resurrect_relationships } to be resurrected"

        return FlowMessage(output_text=msg)

    def step_upload_cdf_relationships(self, flow_msg: FlowMessage = None):
        if not self.cdf_client:
            logging.error("No CDF client available")
            raise Exception("No CDF client available")

        upload_relationships(self.cdf_client, self.categorized_relationships, max_retries=2, retry_delay=4)

    def step_cleanup(self, flow_msg: FlowMessage):
        self.categorized_assets = None
        self.categorized_relationships = None
